from __future__ import absolute_import
from pymic.net_run.weak_sup.wsl_abstract import WSLSegAgent
from pymic.net_run.weak_sup.wsl_em import WSLEntropyMinimization
from pymic.net_run.weak_sup.wsl_gatedcrf import WSLGatedCRF
from pymic.net_run.weak_sup.wsl_mumford_shah import WSLMumfordShah
from pymic.net_run.weak_sup.wsl_tv import WSLTotalVariation
from pymic.net_run.weak_sup.wsl_ustm import WSLUSTM
from pymic.net_run.weak_sup.wsl_dmpls import WSLDMPLS

WSLMethodDict = {'EntropyMinimization': WSLEntropyMinimization,
    'GatedCRF': WSLGatedCRF,
    'MumfordShah': WSLMumfordShah,
    'TotalVariation': WSLTotalVariation,
    'USTM': WSLUSTM,
    'DMPLS': WSLDMPLS}


# -*- coding: utf-8 -*-
from __future__ import print_function, division
import copy
import logging
import time
import logging
import numpy as np
import random
import torch
import torch.nn as nn
import torchvision.transforms as transforms
from datetime import datetime
from random import random
from torch.optim import lr_scheduler
from tensorboardX import SummaryWriter
from pymic.io.nifty_dataset import NiftyDataset
from pymic.loss.seg.util import get_soft_label
from pymic.loss.seg.util import reshape_prediction_and_ground_truth
from pymic.loss.seg.util import get_classwise_dice
from pymic.net_run.infer_func import Inferer
from pymic.net_run.agent_seg import SegmentationAgent
from pymic.transform.trans_dict import TransformDict
from pymic.loss.seg.mse import MAELoss, MSELoss

RegressionLossDict = {
    'MAELoss': MAELoss,
    'MSELoss': MSELoss
    }

class SelfSLSegAgent(SegmentationAgent):
    """
    Abstract class for self-supervised segmentation.

    :param config: (dict) A dictionary containing the configuration.
    :param stage: (str) One of the stage in `train` (default), `inference` or `test`. 

    .. note::

        In the configuration dictionary, in addition to the four sections (`dataset`,
        `network`, `training` and `inference`) used in fully supervised learning, an 
        extra section `semi_supervised_learning` is needed. See :doc:`usage.ssl` for details.
    """
    def __init__(self, config, stage = 'train'):
        super(SelfSLSegAgent, self).__init__(config, stage)
        self.transform_dict  = TransformDict

    def create_loss_calculator(self):
        if(self.loss_dict is None):
            self.loss_dict = RegressionLossDict
        loss_name = self.config['training']['loss_type']
        if isinstance(loss_name, (list, tuple)):
            raise ValueError("Undefined loss function {0:}".format(loss_name))
        elif (loss_name not in self.loss_dict):
            raise ValueError("Undefined loss function {0:}".format(loss_name))
        else:
            loss_param = self.config['training']
            loss_param['loss_softmax'] = False
            base_loss = self.loss_dict[loss_name](self.config['training'])
        if(self.config['training'].get('deep_supervise', False)):
            raise ValueError("Deep supervised loss not implemented for self-supervised learning")
            # weight = self.config['training'].get('deep_supervise_weight', None)
            # mode   = self.config['training'].get('deep_supervise_mode', 2)
            # params = {'deep_supervise_weight': weight, 
            #           'deep_supervise_mode': mode, 
            #           'base_loss':base_loss}
            # self.loss_calculator = DeepSuperviseLoss(params)
        else:
            self.loss_calculator = base_loss

    def training(self):
        iter_valid  = self.config['training']['iter_valid']
        train_loss  = 0
        self.net.train()
        for it in range(iter_valid):
            try:
                data = next(self.trainIter)
            except StopIteration:
                self.trainIter = iter(self.train_loader)
                data = next(self.trainIter)
            # get the inputs
            inputs  = self.convert_tensor_type(data['image'])
            label   = self.convert_tensor_type(data['label'])                 
                   
            # for debug
            # from pymic.io.image_read_write import save_nd_array_as_image
            # for i in range(inputs.shape[0]):
            #     image_i = inputs[i][0]
            #     image_name = "temp/image_{0:}_{1:}.nii.gz".format(it, i)
            #     save_nd_array_as_image(image_i, image_name, reference_name = None)
            # return

            inputs, label = inputs.to(self.device), label.to(self.device)
            
            # zero the parameter gradients
            self.optimizer.zero_grad()
                
            # forward + backward + optimize
            outputs = self.net(inputs)
            outputs = nn.Sigmoid()(outputs)
            loss = self.get_loss_value(data, outputs, label)
            loss.backward()
            self.optimizer.step()
            train_loss = train_loss + loss.item()
            # get dice evaluation for each class
            if(isinstance(outputs, tuple) or isinstance(outputs, list)):
                outputs = outputs[0] 

        train_avg_loss = train_loss / iter_valid
        train_scalers = {'loss': train_avg_loss}
        return train_scalers
        
    def validation(self):
        if(self.inferer is None):
            infer_cfg = self.config['testing']
            self.inferer = Inferer(infer_cfg)
        
        valid_loss_list = []
        validIter  = iter(self.valid_loader)
        with torch.no_grad():
            self.net.eval()
            for data in validIter:
                inputs = self.convert_tensor_type(data['image'])
                label  = self.convert_tensor_type(data['label'])
                inputs, label  = inputs.to(self.device), label.to(self.device)
                outputs = self.inferer.run(self.net, inputs)
                outputs = nn.Sigmoid()(outputs)
                # The tensors are on CPU when calculating loss for validation data
                loss = self.get_loss_value(data, outputs, label)
                valid_loss_list.append(loss.item())

        valid_avg_loss = np.asarray(valid_loss_list).mean()
        valid_scalers = {'loss': valid_avg_loss}
        return valid_scalers

    def write_scalars(self, train_scalars, valid_scalars, lr_value, glob_it):
        loss_scalar ={'train':train_scalars['loss'], 
                      'valid':valid_scalars['loss']}
        self.summ_writer.add_scalars('loss', loss_scalar, glob_it)
        self.summ_writer.add_scalars('lr', {"lr": lr_value}, glob_it)
        logging.info('train loss {0:.4f}'.format(train_scalars['loss']))        
        logging.info('valid loss {0:.4f}'.format(valid_scalars['loss']))  

    def train_valid(self):
        device_ids = self.config['training']['gpus']
        if(len(device_ids) > 1):
            self.device = torch.device("cuda:0")
            self.net = nn.DataParallel(self.net, device_ids = device_ids)
        else:
            self.device = torch.device("cuda:{0:}".format(device_ids[0]))
        self.net.to(self.device)
        ckpt_dir    = self.config['training']['ckpt_save_dir']
        ckpt_prefix = self.config['training'].get('ckpt_prefix', None)
        if(ckpt_prefix is None):
            ckpt_prefix = ckpt_dir.split('/')[-1]
        iter_start  = self.config['training']['iter_start']
        iter_max    = self.config['training']['iter_max']
        iter_valid  = self.config['training']['iter_valid']
        iter_save   = self.config['training'].get('iter_save', None)
        early_stop_it = self.config['training'].get('early_stop_patience', None)
        if(iter_save is None):
            iter_save_list = [iter_max]
        elif(isinstance(iter_save, (tuple, list))):
            iter_save_list = iter_save
        else:
            iter_save_list = range(0, iter_max + 1, iter_save)

        self.min_val_loss = 10000.0
        self.max_val_it   = 0
        self.best_model_wts = None 
        self.checkpoint = None
        if(iter_start > 0):
            checkpoint_file = "{0:}/{1:}_{2:}.pt".format(ckpt_dir, ckpt_prefix, iter_start)
            self.checkpoint = torch.load(checkpoint_file, map_location = self.device)
            # assert(self.checkpoint['iteration'] == iter_start)
            if(len(device_ids) > 1):
                self.net.module.load_state_dict(self.checkpoint['model_state_dict'])
            else:
                self.net.load_state_dict(self.checkpoint['model_state_dict'])
            self.min_val_loss = self.checkpoint.get('valid_loss', 10000)
            # self.max_val_it   = self.checkpoint['iteration']
            self.max_val_it   = iter_start
            self.best_model_wts = self.checkpoint['model_state_dict']
            
        self.create_optimizer(self.get_parameters_to_update())
        self.create_loss_calculator()
    
        self.trainIter  = iter(self.train_loader)
        
        logging.info("{0:} training start".format(str(datetime.now())[:-7]))
        self.summ_writer = SummaryWriter(self.config['training']['ckpt_save_dir'])
        self.glob_it = iter_start
        for it in range(iter_start, iter_max, iter_valid):
            lr_value = self.optimizer.param_groups[0]['lr']
            t0 = time.time()
            train_scalars = self.training()
            t1 = time.time()
            valid_scalars = self.validation()
            t2 = time.time()
            if(isinstance(self.scheduler, lr_scheduler.ReduceLROnPlateau)):
                self.scheduler.step(-valid_scalars['loss'])
            else:
                self.scheduler.step()

            self.glob_it = it + iter_valid
            logging.info("\n{0:} it {1:}".format(str(datetime.now())[:-7], self.glob_it))
            logging.info('learning rate {0:}'.format(lr_value))
            logging.info("training/validation time: {0:.2f}s/{1:.2f}s".format(t1-t0, t2-t1))
            self.write_scalars(train_scalars, valid_scalars, lr_value, self.glob_it)
            if(valid_scalars['loss'] < self.min_val_loss):
                self.min_val_loss = valid_scalars['loss']
                self.max_val_it   = self.glob_it
                if(len(device_ids) > 1):
                    self.best_model_wts = copy.deepcopy(self.net.module.state_dict())
                else:
                    self.best_model_wts = copy.deepcopy(self.net.state_dict())

            stop_now = True if(early_stop_it is not None and \
                self.glob_it - self.max_val_it > early_stop_it) else False
            if ((self.glob_it in iter_save_list) or stop_now):
                save_dict = {'iteration': self.glob_it,
                             'valid_loss': valid_scalars['loss'],
                             'model_state_dict': self.net.module.state_dict() \
                                 if len(device_ids) > 1 else self.net.state_dict(),
                             'optimizer_state_dict': self.optimizer.state_dict()}
                save_name = "{0:}/{1:}_{2:}.pt".format(ckpt_dir, ckpt_prefix, self.glob_it)
                torch.save(save_dict, save_name) 
                txt_file = open("{0:}/{1:}_latest.txt".format(ckpt_dir, ckpt_prefix), 'wt')
                txt_file.write(str(self.glob_it))
                txt_file.close()
            if(stop_now):
                logging.info("The training is early stopped")
                break
        # save the best performing checkpoint
        save_dict = {'iteration': self.max_val_it,
                    'valid_loss': self.min_val_loss,
                    'model_state_dict': self.best_model_wts,
                    'optimizer_state_dict': self.optimizer.state_dict()}
        save_name = "{0:}/{1:}_{2:}.pt".format(ckpt_dir, ckpt_prefix, self.max_val_it)
        torch.save(save_dict, save_name) 
        txt_file = open("{0:}/{1:}_best.txt".format(ckpt_dir, ckpt_prefix), 'wt')
        txt_file.write(str(self.max_val_it))
        txt_file.close()
        logging.info('The best performing iter is {0:}, valid loss {1:}'.format(\
            self.max_val_it, self.min_val_loss))
        self.summ_writer.close()


# -*- coding: utf-8 -*-
"""
Built-in networks for segmentation.

* UNet2D :mod:`pymic.net.net2d.unet2d.UNet2D`
* UNet2D_DualBranch :mod:`pymic.net.net2d.unet2d_dual_branch.UNet2D_DualBranch`
* UNet2D_CCT  :mod:`pymic.net.net2d.unet2d_cct.UNet2D_CCT`
* UNet2D_ScSE  :mod:`pymic.net.net2d.unet2d_scse.UNet2D_ScSE`
* AttentionUNet2D  :mod:`pymic.net.net2d.unet2d_attention.AttentionUNet2D`
* NestedUNet2D :mod:`pymic.net.net2d.unet2d_nest.NestedUNet2D`
* COPLENet  :mod:`pymic.net.net2d.cople_net.COPLENet`
* UNet2D5 :mod:`pymic.net.net3d.unet2d5.UNet2D5`
* UNet3D :mod:`pymic.net.net3d.unet3d.UNet3D`
* UNet3D_ScSE :mod:`pymic.net.net3d.unet3d_scse.UNet3D_ScSE`
"""
from __future__ import print_function, division
from pymic.net.net2d.unet2d import UNet2D
from pymic.net.net2d.unet2d_dual_branch import UNet2D_DualBranch
from pymic.net.net2d.unet2d_cct import UNet2D_CCT
from pymic.net.net2d.cople_net import COPLENet
from pymic.net.net2d.unet2d_attention import AttentionUNet2D
from pymic.net.net2d.unet2d_nest import NestedUNet2D
from pymic.net.net2d.unet2d_scse import UNet2D_ScSE
from pymic.net.net2d.trans2d.transunet import TransUNet
from pymic.net.net2d.trans2d.swinunet import SwinUNet
from pymic.net.net3d.unet2d5 import UNet2D5
from pymic.net.net3d.unet3d import UNet3D
from pymic.net.net3d.unet3d_scse import UNet3D_ScSE
from pymic.net.net3d.unet3d_dual_branch import UNet3D_DualBranch
from pymic.net.net3d.trans3d.unetr import UNETR
from pymic.net.net3d.trans3d.unetr_pp import UNETR_PP
from pymic.net.net3d.trans3d.MedFormer_v1 import MedFormerV1
from pymic.net.net3d.trans3d.MedFormer_v2 import MedFormerV2
from pymic.net.net3d.trans3d.MedFormer_v3 import MedFormerV3
from pymic.net.net3d.trans3d.MedFormer_va1 import MedFormerVA1

SegNetDict = {
	'UNet2D': UNet2D,
	'UNet2D_DualBranch': UNet2D_DualBranch,
	'UNet2D_CCT': UNet2D_CCT,
	'COPLENet': COPLENet,
	'AttentionUNet2D': AttentionUNet2D,
	'NestedUNet2D': NestedUNet2D,
	'UNet2D_ScSE': UNet2D_ScSE,
	'TransUNet': TransUNet,
	'SwinUNet': SwinUNet,
	'UNet2D5': UNet2D5,
	'UNet3D': UNet3D,
	'UNet3D_ScSE': UNet3D_ScSE,
	'UNet3D_DualBranch': UNet3D_DualBranch, 
	'UNETR': UNETR,
	'UNETR_PP': UNETR_PP,
	'MedFormerV1': MedFormerV1,
	'MedFormerV2': MedFormerV2,
	'MedFormerV3': MedFormerV3,
	'MedFormerVA1':MedFormerVA1
	}



import torch
import torch.nn as nn
from pymic.loss.seg.abstract import AbstractSegLoss

class MSELoss(AbstractSegLoss):
    """
    Mean Sequare Loss for segmentation tasks.
    The parameters should be written in the `params` dictionary, and it has the
    following fields:

    :param `loss_softmax`: (bool) Apply softmax to the prediction of network or not. 
    """
    def __init__(self, params = None):
        super(MSELoss, self).__init__(params)
            
    def forward(self, loss_input_dict):
        predict = loss_input_dict['prediction']
        soft_y  = loss_input_dict['ground_truth']
        
        if(isinstance(predict, (list, tuple))):
            predict = predict[0]
        if(self.softmax):
            predict = nn.Softmax(dim = 1)(predict)
        mse  = torch.square(predict - soft_y)
        mse  = torch.mean(mse) 
        return mse 
    
    
class MAELoss(AbstractSegLoss):
    """
    Mean Absolute Loss for segmentation tasks.
    The arguments should be written in the `params` dictionary, and it has the
    following fields:

    :param `loss_softmax`: (bool) Apply softmax to the prediction of network or not. 
    """
    def __init__(self, params = None):
        super(MAELoss, self).__init__(params)
    
    def forward(self, loss_input_dict):
        predict = loss_input_dict['prediction']
        soft_y  = loss_input_dict['ground_truth']
        
        if(isinstance(predict, (list, tuple))):
            predict = predict[0]
        if(self.softmax):
            predict = nn.Softmax(dim = 1)(predict)
        mae = torch.abs(predict - soft_y)
        mae = torch.mean(mae)
        return mae 



# -*- coding: utf-8 -*-
from __future__ import print_function, division
import logging
import numpy as np
import torch
from pymic.loss.seg.util import get_soft_label
from pymic.loss.seg.util import reshape_prediction_and_ground_truth
from pymic.loss.seg.util import get_classwise_dice
from pymic.loss.seg.ssl import EntropyLoss
from pymic.net_run.semi_sup import SSLSegAgent
from pymic.transform.trans_dict import TransformDict
from pymic.util.ramps import get_rampup_ratio

class SSLEntropyMinimization(SSLSegAgent):
    """
    Using Entropy Minimization for semi-supervised segmentation. 

    * Reference: Yves Grandvalet and Yoshua Bengio:
      Semi-supervised Learningby Entropy Minimization.
      `NeurIPS, 2005. <https://papers.nips.cc/paper/2004/file/96f2b50b5d3613adf9c27049b2a888c7-Paper.pdf>`_ 
    
    :param config: (dict) A dictionary containing the configuration.
    :param stage: (str) One of the stage in `train` (default), `inference` or `test`. 

    .. note::

        In the configuration dictionary, in addition to the four sections (`dataset`,
        `network`, `training` and `inference`) used in fully supervised learning, an 
        extra section `semi_supervised_learning` is needed. See :doc:`usage.ssl` for details.
    """
    def __init__(self, config, stage = 'train'):
        super(SSLEntropyMinimization, self).__init__(config, stage)
        self.transform_dict  = TransformDict
        self.train_set_unlab = None 

    def training(self):
        class_num   = self.config['network']['class_num']
        iter_valid  = self.config['training']['iter_valid']
        ssl_cfg     = self.config['semi_supervised_learning']
        iter_max     = self.config['training']['iter_max']
        rampup_start = ssl_cfg.get('rampup_start', 0)
        rampup_end   = ssl_cfg.get('rampup_end', iter_max)
        train_loss  = 0
        train_loss_sup = 0
        train_loss_reg = 0
        train_dice_list = []
        self.net.train()
        for it in range(iter_valid):
            try:
                data_lab = next(self.trainIter)
            except StopIteration:
                self.trainIter = iter(self.train_loader)
                data_lab = next(self.trainIter)
            try:
                data_unlab = next(self.trainIter_unlab)
            except StopIteration:
                self.trainIter_unlab = iter(self.train_loader_unlab)
                data_unlab = next(self.trainIter_unlab)

            # get the inputs
            x0   = self.convert_tensor_type(data_lab['image'])
            y0   = self.convert_tensor_type(data_lab['label_prob'])  
            x1   = self.convert_tensor_type(data_unlab['image'])
            inputs = torch.cat([x0, x1], dim = 0)               
            inputs, y0 = inputs.to(self.device), y0.to(self.device)
            
            # zero the parameter gradients
            self.optimizer.zero_grad()
                
            # forward + backward + optimize
            outputs = self.net(inputs)
            n0 = list(x0.shape)[0] 
            p0 = outputs[:n0]
            loss_sup = self.get_loss_value(data_lab, p0, y0)
            loss_dict = {"prediction":outputs, 'softmax':True}
            loss_reg  = EntropyLoss()(loss_dict)
            
            rampup_ratio = get_rampup_ratio(self.glob_it, rampup_start, rampup_end, "sigmoid")
            regular_w = ssl_cfg.get('regularize_w', 0.1) * rampup_ratio

            loss = loss_sup + regular_w*loss_reg
            # if (self.config['training']['use'])
            loss.backward()
            self.optimizer.step()

            train_loss = train_loss + loss.item()
            train_loss_sup = train_loss_sup + loss_sup.item()
            train_loss_reg = train_loss_reg + loss_reg.item() 
            # get dice evaluation for each class in annotated images
            if(isinstance(p0, tuple) or isinstance(p0, list)):
                p0 = p0[0] 
            p0_argmax = torch.argmax(p0, dim = 1, keepdim = True)
            p0_soft   = get_soft_label(p0_argmax, class_num, self.tensor_type)
            p0_soft, y0 = reshape_prediction_and_ground_truth(p0_soft, y0) 
            dice_list   = get_classwise_dice(p0_soft, y0)
            train_dice_list.append(dice_list.cpu().numpy())
        train_avg_loss = train_loss / iter_valid
        train_avg_loss_sup = train_loss_sup / iter_valid
        train_avg_loss_reg = train_loss_reg / iter_valid
        train_cls_dice = np.asarray(train_dice_list).mean(axis = 0)
        train_avg_dice = train_cls_dice[1:].mean()

        train_scalers = {'loss': train_avg_loss, 'loss_sup':train_avg_loss_sup,
            'loss_reg':train_avg_loss_reg, 'regular_w':regular_w,
            'avg_fg_dice':train_avg_dice,     'class_dice': train_cls_dice}
        return train_scalers


# -*- coding: utf-8 -*-
from __future__ import print_function, division

import torch
import torch.nn as nn

class AbstractClassificationLoss(nn.Module):
    """
    Abstract Classification Loss.
    """
    def __init__(self, params = None):
        super(AbstractClassificationLoss, self).__init__()
    
    def forward(self, loss_input_dict):
        """
        The arguments should be written in the `loss_input_dict` dictionary, and it has the
        following fields. 
        
        :param prediction: A prediction with shape of [N, C] where C is the class number.
        :param ground_truth: The corresponding ground truth, with shape of [N, 1].

        Note that `prediction` is the digit output of a network, before using softmax.
        """
        pass

class CrossEntropyLoss(AbstractClassificationLoss):
    """
    Standard Softmax-based  CE loss.
    """
    def __init__(self, params = None):
        super(CrossEntropyLoss, self).__init__(params)
        self.ce_loss = nn.CrossEntropyLoss()
    
    def forward(self, loss_input_dict):
        predict = loss_input_dict['prediction']
        labels  = loss_input_dict['ground_truth']
        loss = self.ce_loss(predict, labels)
        return loss

class SigmoidCELoss(AbstractClassificationLoss):
    """
    Sigmoid-based CE loss.
    """
    def __init__(self, params = None):
        super(SigmoidCELoss, self).__init__(params)
    
    def forward(self, loss_input_dict):
        predict = loss_input_dict['prediction']
        labels  = loss_input_dict['ground_truth']
        predict = nn.Sigmoid()(predict) * 0.999 + 5e-4
        loss = - labels * torch.log(predict) - (1 - labels) * torch.log( 1 - predict)
        loss = loss.mean()
        return loss

class L1Loss(AbstractClassificationLoss):
    """
    L1 (MAE) loss for classification
    """
    def __init__(self, params = None):
        super(L1Loss, self).__init__(params)
        self.l1_loss = nn.L1Loss()
    
    def forward(self, loss_input_dict):
        predict = loss_input_dict['prediction']
        labels  = loss_input_dict['ground_truth'][:, None] # reshape to N, 1
        softmax = nn.Softmax(dim = 1)
        predict = softmax(predict)
        loss = self.l1_loss(predict, labels)
        return loss

class MSELoss(AbstractClassificationLoss):
    """
    Mean Square Error loss for classification.
    """
    def __init__(self, params = None):
        super(MSELoss, self).__init__(params)
        self.mse_loss = nn.MSELoss()
    
    def forward(self, loss_input_dict):
        predict = loss_input_dict['prediction']
        labels  = loss_input_dict['ground_truth'][:, None] # reshape to N, 1
        softmax = nn.Softmax(dim = 1)
        predict = softmax(predict)
        loss = self.mse_loss(predict, labels)
        return loss

class NLLLoss(AbstractClassificationLoss):
    """
    The negative log likelihood loss for classification.
    """
    def __init__(self, params = None):
        super(NLLLoss, self).__init__(params)
        self.nll_loss = nn.NLLLoss()
    
    def forward(self, loss_input_dict):
        predict = loss_input_dict['prediction']
        labels  = loss_input_dict['ground_truth']
        logsoft = nn.LogSoftmax(dim = 1)
        predict = logsoft(predict)
        loss = self.nll_loss(predict, labels)
        return loss


# pretrained models from pytorch: https://pytorch.org/vision/0.8/models.html
from __future__ import print_function, division

import itertools
import torch
import torch.nn as nn
import torchvision.models as models

# buildin_net_dict={
#     'resnet18': models.resnet18,
#     'alexnet':  models.alexnet,
#     'vgg16':    models.vgg16,
#     'squeezenet': models.squeezenet1_0,
#     'densenet': models.densenet161,
#     'inception':models.inception_v3,
#     'googlenet':models.googlenet,
#     'shufflenet': models.shufflenet_v2_x1_0,
#     'mobilenet':models.mobilenet,
#     'resnext50':models.resnext50_32x4d,
#     'wide_resnet': models.wide_resnet50_2,
#     'mnasnet': models.mnasnet1_0
# }

class BuiltInNet(nn.Module):
    """
    Built-in Network in Pytorch for classification.
    Parameters should be set in the `params` dictionary that contains the 
    following fields:

    :param input_chns: (int) Input channel number, default is 3.
    :param pretrain: (bool) Using pretrained model or not, default is True. 
    :param update_mode: (str) The strategy for updating layers: "`all`" means updating
        all the layers, and "`last`" (by default) means updating the last layer, 
        as well as the first layer when `input_chns` is not 3.
    """
    def __init__(self, params):
        super(BuiltInNet, self).__init__()
        self.params   = params
        self.in_chns  = params.get('input_chns', 3)
        self.pretrain = params.get('pretrain', True)
        self.update_mode = params.get('update_mode', "last")
        self.net = None 
    
    def forward(self, x):
        return self.net(x)

    def get_parameters_to_update(self):
        pass

class ResNet18(BuiltInNet):
    """
    ResNet18 for classification.
    Parameters should be set in the `params` dictionary that contains the 
    following fields:

    :param input_chns: (int) Input channel number, default is 3.
    :param pretrain: (bool) Using pretrained model or not, default is True. 
    :param update_mode: (str) The strategy for updating layers: "`all`" means updating
        all the layers, and "`last`" (by default) means updating the last layer, 
        as well as the first layer when `input_chns` is not 3.
    """
    def __init__(self, params):
        super(ResNet18, self).__init__(params)
        self.net = models.resnet18(pretrained = self.pretrain)
        
        # replace the last layer 
        num_ftrs = self.net.fc.in_features
        self.net.fc = nn.Linear(num_ftrs, params['class_num'])

        # replace the first layer when in_chns is not 3
        if(self.in_chns != 3):
            self.net.conv1 = nn.Conv2d(self.in_chns, 64, kernel_size=(7, 7), 
                stride=(2, 2), padding=(3, 3), bias=False)
    
    def get_parameters_to_update(self):
        if(self.update_mode == "all"):
            return self.net.parameters()
        elif(self.update_mode == "last"):
            params = self.net.fc.parameters()
            if(self.in_chns !=3):
                # combining the two iterables into a single one 
                # see: https://dzone.com/articles/python-joining-multiple
                params = itertools.chain()
                for pram in [self.net.fc.parameters(), self.net.conv1.parameters()]:
                    params = itertools.chain(params, pram)
            return  params
        else:
            raise(ValueError("update_mode can only be 'all' or 'last'."))

class VGG16(BuiltInNet):
    """
    VGG16 for classification.
    Parameters should be set in the `params` dictionary that contains the 
    following fields:

    :param input_chns: (int) Input channel number, default is 3.
    :param pretrain: (bool) Using pretrained model or not, default is True. 
    :param update_mode: (str) The strategy for updating layers: "`all`" means updating
        all the layers, and "`last`" (by default) means updating the last layer, 
        as well as the first layer when `input_chns` is not 3.
    """
    def __init__(self, params):
        super(VGG16, self).__init__(params)
        self.net = models.vgg16(pretrained = self.pretrain)
        
        # replace the last layer 
        num_ftrs = self.net.classifier[-1].in_features
        self.net.classifier[-1] = nn.Linear(num_ftrs, params['class_num'])

        # replace the first layer when in_chns is not 3
        if(self.in_chns != 3):
            self.net.features[0] = nn.Conv2d(self.in_chns, 64, kernel_size=(3, 3), 
                stride=(1, 1), padding=(1, 1), bias=False)
    
    def get_parameters_to_update(self):
        if(self.update_mode == "all"):
            return self.net.parameters()
        elif(self.update_mode == "last"):
            params = self.net.classifier[-1].parameters()
            if(self.in_chns !=3):
                params = itertools.chain()
                for pram in [self.net.classifier[-1].parameters(), self.net.features[0].parameters()]:
                    params = itertools.chain(params, pram)
            return  params
        else:
            raise(ValueError("update_mode can only be 'all' or 'last'."))

class MobileNetV2(BuiltInNet):
    """
    MobileNetV2 for classification.
    Parameters should be set in the `params` dictionary that contains the 
    following fields:

    :param input_chns: (int) Input channel number, default is 3.
    :param pretrain: (bool) Using pretrained model or not, default is True. 
    :param update_mode: (str) The strategy for updating layers: "`all`" means updating
        all the layers, and "`last`" (by default) means updating the last layer, 
        as well as the first layer when `input_chns` is not 3.
    """
    def __init__(self, params):
        super(MobileNetV2, self).__init__(params)
        self.net = models.mobilenet_v2(pretrained = self.pretrain)
        
        # replace the last layer 
        num_ftrs = self.net.last_channel
        self.net.classifier[-1] = nn.Linear(num_ftrs, params['class_num'])

        # replace the first layer when in_chns is not 3
        if(self.in_chns != 3):
            self.net.features[0][0] = nn.Conv2d(self.in_chns, 32, kernel_size=(3, 3), 
                stride=(2, 2), padding=(1, 1), bias=False)
    
    def get_parameters_to_update(self):
        if(self.update_mode == "all"):
            return self.net.parameters()
        elif(self.update_mode == "last"):
            params = self.net.classifier[-1].parameters()
            if(self.in_chns !=3):
                params = itertools.chain()
                for pram in [self.net.classifier[-1].parameters(), self.net.features[0][0].parameters()]:
                    params = itertools.chain(params, pram)
            return  params
        else:
            raise(ValueError("update_mode can only be 'all' or 'last'."))

if __name__ == "__main__":
    params = {"class_num": 2, "pretrain": False, "input_chns": 3}
    net = ResNet18(params)
    print(net)


# -*- coding: utf-8 -*-
from __future__ import print_function, division

import numpy as np 
import torch
import torch.nn as nn
import torch.nn.functional as F
from typing import Optional, Sequence, Tuple, Union
from pymic.net.net3d.trans3d.unetr_pp_block import UnetOutBlock, UnetResBlock, get_conv_layer
from timm.models.layers import trunc_normal_
from monai.utils import optional_import
from monai.networks.blocks.convolutions import Convolution
from monai.networks.layers.factories import Act, Norm
from monai.networks.layers.utils import get_act_layer, get_norm_layer

einops, _ = optional_import("einops")

class LayerNorm(nn.Module):
    def __init__(self, normalized_shape, eps=1e-6, data_format="channels_last"):
        super().__init__()
        self.weight = nn.Parameter(torch.ones(normalized_shape))
        self.bias = nn.Parameter(torch.zeros(normalized_shape))
        self.eps = eps
        self.data_format = data_format
        if self.data_format not in ["channels_last", "channels_first"]:
            raise NotImplementedError
        self.normalized_shape = (normalized_shape,)

    def forward(self, x):
        if self.data_format == "channels_last":
            return F.layer_norm(x, self.normalized_shape, self.weight, self.bias, self.eps)
        elif self.data_format == "channels_first":
            u = x.mean(1, keepdim=True)
            s = (x - u).pow(2).mean(1, keepdim=True)
            x = (x - u) / torch.sqrt(s + self.eps)
            x = self.weight[:, None, None] * x + self.bias[:, None, None]
            return x

class EPA(nn.Module):
    """
        Efficient Paired Attention Block, based on: "Shaker et al.,
        UNETR++: Delving into Efficient and Accurate 3D Medical Image Segmentation"
        """
    def __init__(self, input_size, hidden_size, proj_size, num_heads=4, qkv_bias=False,
                 channel_attn_drop=0.1, spatial_attn_drop=0.1):
        super().__init__()
        self.num_heads = num_heads
        self.temperature = nn.Parameter(torch.ones(num_heads, 1, 1))
        self.temperature2 = nn.Parameter(torch.ones(num_heads, 1, 1))

        # qkvv are 4 linear layers (query_shared, key_shared, value_spatial, value_channel)
        self.qkvv = nn.Linear(hidden_size, hidden_size * 4, bias=qkv_bias)

        # E and F are projection matrices with shared weights used in spatial attention module to project
        # keys and values from HWD-dimension to P-dimension
        self.E = self.F = nn.Linear(input_size, proj_size)

        self.attn_drop = nn.Dropout(channel_attn_drop)
        self.attn_drop_2 = nn.Dropout(spatial_attn_drop)

        self.out_proj = nn.Linear(hidden_size, int(hidden_size // 2))
        self.out_proj2 = nn.Linear(hidden_size, int(hidden_size // 2))

    def forward(self, x):
        B, N, C = x.shape

        qkvv = self.qkvv(x).reshape(B, N, 4, self.num_heads, C // self.num_heads)

        qkvv = qkvv.permute(2, 0, 3, 1, 4)

        q_shared, k_shared, v_CA, v_SA = qkvv[0], qkvv[1], qkvv[2], qkvv[3]

        q_shared = q_shared.transpose(-2, -1)
        k_shared = k_shared.transpose(-2, -1)
        v_CA = v_CA.transpose(-2, -1)
        v_SA = v_SA.transpose(-2, -1)

        k_shared_projected = self.E(k_shared)

        v_SA_projected = self.F(v_SA)

        q_shared = torch.nn.functional.normalize(q_shared, dim=-1)
        k_shared = torch.nn.functional.normalize(k_shared, dim=-1)

        attn_CA = (q_shared @ k_shared.transpose(-2, -1)) * self.temperature

        attn_CA = attn_CA.softmax(dim=-1)
        attn_CA = self.attn_drop(attn_CA)

        x_CA = (attn_CA @ v_CA).permute(0, 3, 1, 2).reshape(B, N, C)

        attn_SA = (q_shared.permute(0, 1, 3, 2) @ k_shared_projected) * self.temperature2

        attn_SA = attn_SA.softmax(dim=-1)
        attn_SA = self.attn_drop_2(attn_SA)

        x_SA = (attn_SA @ v_SA_projected.transpose(-2, -1)).permute(0, 3, 1, 2).reshape(B, N, C)

        # Concat fusion
        x_SA = self.out_proj(x_SA)
        x_CA = self.out_proj2(x_CA)
        x = torch.cat((x_SA, x_CA), dim=-1)
        return x

    @torch.jit.ignore
    def no_weight_decay(self):
        return {'temperature', 'temperature2'}


class TransformerBlock(nn.Module):
    """
    A transformer block, based on: "Shaker et al.,
    UNETR++: Delving into Efficient and Accurate 3D Medical Image Segmentation"
    """

    def __init__(
            self,
            input_size: int,
            hidden_size: int,
            proj_size: int,
            num_heads: int,
            dropout_rate: float = 0.0,
            pos_embed=False,
    ) -> None:
        """
        Args:
            input_size: the size of the input for each stage.
            hidden_size: dimension of hidden layer.
            proj_size: projection size for keys and values in the spatial attention module.
            num_heads: number of attention heads.
            dropout_rate: faction of the input units to drop.
            pos_embed: bool argument to determine if positional embedding is used.

        """

        super().__init__()

        if not (0 <= dropout_rate <= 1):
            raise ValueError("dropout_rate should be between 0 and 1.")

        if hidden_size % num_heads != 0:
            print("Hidden size is ", hidden_size)
            print("Num heads is ", num_heads)
            raise ValueError("hidden_size should be divisible by num_heads.")

        self.norm = nn.LayerNorm(hidden_size)
        self.gamma = nn.Parameter(1e-6 * torch.ones(hidden_size), requires_grad=True)
        self.epa_block = EPA(input_size=input_size, hidden_size=hidden_size, proj_size=proj_size, num_heads=num_heads, channel_attn_drop=dropout_rate,spatial_attn_drop=dropout_rate)
        self.conv51 = UnetResBlock(3, hidden_size, hidden_size, kernel_size=3, stride=1, norm_name="batch")
        self.conv8 = nn.Sequential(nn.Dropout3d(0.1, False), nn.Conv3d(hidden_size, hidden_size, 1))

        self.pos_embed = None
        if pos_embed:
            self.pos_embed = nn.Parameter(torch.zeros(1, input_size, hidden_size))

    def forward(self, x):
        B, C, H, W, D = x.shape

        x = x.reshape(B, C, H * W * D).permute(0, 2, 1)

        if self.pos_embed is not None:
            x = x + self.pos_embed
        attn = x + self.gamma * self.epa_block(self.norm(x))

        attn_skip = attn.reshape(B, H, W, D, C).permute(0, 4, 1, 2, 3)  # (B, C, H, W, D)
        attn = self.conv51(attn_skip)
        x = attn_skip + self.conv8(attn)

        return x

class UnetrPPEncoder(nn.Module):
    def __init__(self, input_size=[32 * 32 * 32, 16 * 16 * 16, 8 * 8 * 8, 4 * 4 * 4],dims=[32, 64, 128, 256],
                 proj_size =[64,64,64,32], depths=[3, 3, 3, 3],  num_heads=4, spatial_dims=3, in_channels=1, dropout=0.0, transformer_dropout_rate=0.15 ,**kwargs):
        super().__init__()

        self.downsample_layers = nn.ModuleList()  # stem and 3 intermediate downsampling conv layers
        stem_layer = nn.Sequential(
            get_conv_layer(spatial_dims, in_channels, dims[0], kernel_size=(2, 4, 4), stride=(2, 4, 4),
                           dropout=dropout, conv_only=True, ),
            get_norm_layer(name=("group", {"num_groups": in_channels}), channels=dims[0]),
        )
        self.downsample_layers.append(stem_layer)
        for i in range(3):
            downsample_layer = nn.Sequential(
                get_conv_layer(spatial_dims, dims[i], dims[i + 1], kernel_size=(2, 2, 2), stride=(2, 2, 2),
                               dropout=dropout, conv_only=True, ),
                get_norm_layer(name=("group", {"num_groups": dims[i]}), channels=dims[i + 1]),
            )
            self.downsample_layers.append(downsample_layer)

        self.stages = nn.ModuleList()  # 4 feature resolution stages, each consisting of multiple Transformer blocks
        for i in range(4):
            stage_blocks = []
            for j in range(depths[i]):
                stage_blocks.append(TransformerBlock(input_size=input_size[i], hidden_size=dims[i],  proj_size=proj_size[i], num_heads=num_heads,
                                     dropout_rate=transformer_dropout_rate, pos_embed=True))
            self.stages.append(nn.Sequential(*stage_blocks))
        self.hidden_states = []
        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, (nn.Conv2d, nn.Linear)):
            trunc_normal_(m.weight, std=.02)
            if m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, (LayerNorm, nn.LayerNorm)):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    def forward_features(self, x):
        hidden_states = []

        x = self.downsample_layers[0](x)
        x = self.stages[0](x)

        hidden_states.append(x)

        for i in range(1, 4):
            x = self.downsample_layers[i](x)
            x = self.stages[i](x)
            if i == 3:  # Reshape the output of the last stage
                x = einops.rearrange(x, "b c h w d -> b (h w d) c")
            hidden_states.append(x)
        return x, hidden_states

    def forward(self, x):
        x, hidden_states = self.forward_features(x)
        return x, hidden_states


class UnetrUpBlock(nn.Module):
    def     __init__(
            self,
            spatial_dims: int,
            in_channels: int,
            out_channels: int,
            kernel_size: Union[Sequence[int], int],
            upsample_kernel_size: Union[Sequence[int], int],
            norm_name: Union[Tuple, str],
            proj_size: int = 64,
            num_heads: int = 4,
            out_size: int = 0,
            depth: int = 3,
            conv_decoder: bool = False,
    ) -> None:
        """
        Args:
            spatial_dims: number of spatial dimensions.
            in_channels: number of input channels.
            out_channels: number of output channels.
            kernel_size: convolution kernel size.
            upsample_kernel_size: convolution kernel size for transposed convolution layers.
            norm_name: feature normalization type and arguments.
            proj_size: projection size for keys and values in the spatial attention module.
            num_heads: number of heads inside each EPA module.
            out_size: spatial size for each decoder.
            depth: number of blocks for the current decoder stage.
        """

        super().__init__()
        upsample_stride = upsample_kernel_size
        self.transp_conv = get_conv_layer(
            spatial_dims,
            in_channels,
            out_channels,
            kernel_size=upsample_kernel_size,
            stride=upsample_stride,
            conv_only=True,
            is_transposed=True,
        )

        # 4 feature resolution stages, each consisting of multiple residual blocks
        self.decoder_block = nn.ModuleList()

        # If this is the last decoder, use ConvBlock(UnetResBlock) instead of EPA_Block (see suppl. material in the paper)
        if conv_decoder == True:
            self.decoder_block.append(
                UnetResBlock(spatial_dims, out_channels, out_channels, kernel_size=kernel_size, stride=1,
                             norm_name=norm_name, ))
        else:
            stage_blocks = []
            for j in range(depth):
                stage_blocks.append(TransformerBlock(input_size=out_size, hidden_size= out_channels, proj_size=proj_size, num_heads=num_heads,
                                                     dropout_rate=0.15, pos_embed=True))
            self.decoder_block.append(nn.Sequential(*stage_blocks))

    def _init_weights(self, m):
        if isinstance(m, (nn.Conv2d, nn.Linear)):
            trunc_normal_(m.weight, std=.02)
            if m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, (nn.LayerNorm)):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    def forward(self, inp, skip):

        out = self.transp_conv(inp)
        out = out + skip
        out = self.decoder_block[0](out)

        return out


class UNETR_PP(nn.Module):
    """
    UNETR++ based on: "Shaker et al.,
    UNETR++: Delving into Efficient and Accurate 3D Medical Image Segmentation"
    """

    def __init__(self, params):
        """
        Args:
            in_channels: dimension of input channels.
            out_channels: dimension of output channels.
            img_size: dimension of input image.
            feature_size: dimension of network feature size.
            hidden_size: dimension of  the last encoder.
            num_heads: number of attention heads.
            pos_embed: position embedding layer type.
            norm_name: feature normalization type and arguments.
            dropout_rate: faction of the input units to drop.
            depths: number of blocks for each stage.
            dims: number of channel maps for the stages.
            conv_op: type of convolution operation.
            do_ds: use deep supervision to compute the loss.

        """
        super().__init__()
        in_channels  = params['in_chns']
        out_channels = params['class_num']
        img_size     = params['img_size']
        feature_size = params.get('feature_size', 16)
        hidden_size  = params.get('hidden_size', 256)
        num_heads    = params.get('num_heads', 4)
        pos_embed    = params.get('pos_embed', "perceptron")
        norm_name    = params.get('norm_name', "instance")
        dropout_rate = params.get('dropout_rate', 0.0)
        depths       = params.get('depths', [3, 3, 3, 3])
        dims         = params.get('dims', [32, 64, 128, 256])
        conv_op      = nn.Conv3d
        do_ds        = params.get('deep_supervise', True)

        self.do_ds = do_ds
        self.conv_op = conv_op
        self.num_classes = out_channels
        if not (0 <= dropout_rate <= 1):
            raise AssertionError("dropout_rate should be between 0 and 1.")

        if pos_embed not in ["conv", "perceptron"]:
            raise KeyError(f"Position embedding layer of type {pos_embed} is not supported.")

        self.patch_size = (2, 4, 4)
        self.feat_size = (
            img_size[0] // self.patch_size[0] // 8,  # 8 is the downsampling happened through the four encoders stages
            img_size[1] // self.patch_size[1] // 8,  # 8 is the downsampling happened through the four encoders stages
            img_size[2] // self.patch_size[2] // 8,  # 8 is the downsampling happened through the four encoders stages
        )
        self.hidden_size = hidden_size

        self.unetr_pp_encoder = UnetrPPEncoder(dims=dims, depths=depths, num_heads=num_heads)

        self.encoder1 = UnetResBlock(
            spatial_dims=3,
            in_channels=in_channels,
            out_channels=feature_size,
            kernel_size=3,
            stride=1,
            norm_name=norm_name,
        )
        self.decoder5 = UnetrUpBlock(
            spatial_dims=3,
            in_channels=feature_size * 16,
            out_channels=feature_size * 8,
            kernel_size=3,
            upsample_kernel_size=2,
            norm_name=norm_name,
            out_size=8 * 8 * 8,
        )
        self.decoder4 = UnetrUpBlock(
            spatial_dims=3,
            in_channels=feature_size * 8,
            out_channels=feature_size * 4,
            kernel_size=3,
            upsample_kernel_size=2,
            norm_name=norm_name,
            out_size=16 * 16 * 16,
        )
        self.decoder3 = UnetrUpBlock(
            spatial_dims=3,
            in_channels=feature_size * 4,
            out_channels=feature_size * 2,
            kernel_size=3,
            upsample_kernel_size=2,
            norm_name=norm_name,
            out_size=32 * 32 * 32,
        )
        self.decoder2 = UnetrUpBlock(
            spatial_dims=3,
            in_channels=feature_size * 2,
            out_channels=feature_size,
            kernel_size=3,
            upsample_kernel_size=(2, 4, 4),
            norm_name=norm_name,
            out_size=64 * 128 * 128,
            conv_decoder=True,
        )
        self.out1 = UnetOutBlock(spatial_dims=3, in_channels=feature_size, out_channels=out_channels)
        if self.do_ds:
            self.out2 = UnetOutBlock(spatial_dims=3, in_channels=feature_size * 2, out_channels=out_channels)
            self.out3 = UnetOutBlock(spatial_dims=3, in_channels=feature_size * 4, out_channels=out_channels)

    def proj_feat(self, x, hidden_size, feat_size):
        x = x.view(x.size(0), feat_size[0], feat_size[1], feat_size[2], hidden_size)
        x = x.permute(0, 4, 1, 2, 3).contiguous()
        return x

    def forward(self, x_in):
        x_output, hidden_states = self.unetr_pp_encoder(x_in)

        convBlock = self.encoder1(x_in)

        # Four encoders
        enc1 = hidden_states[0]
        enc2 = hidden_states[1]
        enc3 = hidden_states[2]
        enc4 = hidden_states[3]

        # Four decoders
        dec4 = self.proj_feat(enc4, self.hidden_size, self.feat_size)
        dec3 = self.decoder5(dec4, enc3)
        dec2 = self.decoder4(dec3, enc2)
        dec1 = self.decoder3(dec2, enc1)

        out = self.decoder2(dec1, convBlock)
        if self.do_ds:
            logits = [self.out1(out), self.out2(dec1), self.out3(dec2)]
        else:
            logits = self.out1(out)

        return logits


if __name__ == "__main__":
    params = {'in_chns': 1,
              'class_num': 2,
              'img_size': [64, 128, 128]
              }
    net = UNETR_PP(params)
    net.double()

    x  = np.random.rand(2, 1, 64, 128, 128)
    xt = torch.from_numpy(x)
    xt = torch.tensor(xt)
    
    y = net(xt)
    print(len(y))
    for yi in y:
        yi = yi.detach().numpy()
        print(yi.shape)


# -*- coding: utf-8 -*-
from __future__ import print_function, division

import torch
import json
import math
import random
import numpy as np
from scipy import ndimage
from pymic.transform.abstract_transform import AbstractTransform
from pymic.util.image_process import *


class ChannelWiseThreshold(AbstractTransform):
    """
    Thresholding the image for given channels.

    The arguments should be written in the `params` dictionary, and it has the
    following fields:

    :param `ChannelWiseThreshold_channels`: (list/tuple or None) 
        A list of specified channels for thresholding. If None (by default), 
        all the channels will be thresholded.
    :param `ChannelWiseThreshold_threshold_lower`: (list/tuple or None) 
        The lower threshold for the given channels.
    :param `ChannelWiseThreshold_threshold_upper`: (list/tuple or None) 
        The upper threshold for the given channels.  
    :param `ChannelWiseThreshold_replace_lower`: (list/tuple or None) 
        The output value for pixels with an input value lower than the threshold_lower.
    :param `ChannelWiseThreshold_replace_upper`: (list/tuple or None) 
        The output value for pixels with an input value higher than the threshold_upper.      
    :param `ChannelWiseThreshold_inverse`: (optional, bool) 
        Is inverse transform needed for inference. Default is `False`.
    """
    def __init__(self, params):
        super(ChannelWiseThreshold, self).__init__(params)
        self.channels = params['ChannelWiseThreshold_channels'.lower()]
        self.threshold_lower = params['ChannelWiseThreshold_threshold_lower'.lower()]
        self.threshold_upper = params['ChannelWiseThreshold_threshold_upper'.lower()]
        self.replace_lower   = params['ChannelWiseThreshold_replace_lower'.lower()]
        self.replace_upper   = params['ChannelWiseThreshold_replace_upper'.lower()]
        self.inverse = params.get('ChannelWiseThreshold_inverse'.lower(), False)

    def __call__(self, sample):
        image= sample['image']
        channels = range(image.shape[0]) if self.channels is None else self.channels
        for i in range(len(channels)):
            chn = channels[i]
            if((self.threshold_lower is not None) and (self.threshold_lower[i] is not None)):
                t_lower = self.threshold_lower[i]
                r_lower = self.threshold_lower[i]
                if((self.replace_lower is not None) and (self.replace_lower[i] is not None)):
                    r_lower = self.replace_lower[i]
                image[chn][image[chn] < t_lower] = r_lower
            
            if((self.threshold_upper is not None) and (self.threshold_upper[i] is not None)):
                t_upper = self.threshold_upper[i]
                r_upper = self.threshold_upper[i]
                if((self.replace_upper is not None) and (self.replace_upper[i] is not None)):
                    r_upper= self.replace_upper[i]
                image[chn][image[chn] > t_upper] = r_upper
        sample['image'] = image
        return sample

class ChannelWiseThresholdWithNormalize(AbstractTransform):
    """
    Apply thresholding and normalization for given channels. 
    Pixel intensity will be truncated to the range of (lower, upper) and then 
    normalized. If mean_std_mode is True, the mean and std values for pixel
    in the target range is calculated for normalization, and input intensity 
    outside that range will be replaced by random values. Otherwise, the intensity
    will be normalized to [0, 1].
    
    The arguments should be written in the `params` dictionary, and it has the
    following fields:

    :param `ChannelWiseThresholdWithNormalize_channels`: (list/tuple or None) 
        A list of specified channels for thresholding. If None (by default), 
        all the channels will be affected by this transform.
    :param `ChannelWiseThresholdWithNormalize_threshold_lower`: (list/tuple or None) 
        The lower threshold for the given channels.
    :param `ChannelWiseThresholdWithNormalize_threshold_upper`: (list/tuple or None) 
        The upper threshold for the given channels.  
    :param `ChannelWiseThresholdWithNormalize_mean_std_mode`: (bool) 
        If True, using mean and std for normalization. If False, using min and max 
        values for normalization.      
    :param `ChannelWiseThresholdWithNormalize_inverse`: (optional, bool) 
        Is inverse transform needed for inference. Default is `False`.
    """
    def __init__(self, params):
        super(ChannelWiseThresholdWithNormalize, self).__init__(params)
        self.channels = params['ChannelWiseThresholdWithNormalize_channels'.lower()]
        self.threshold_lower = params['ChannelWiseThresholdWithNormalize_threshold_lower'.lower()]
        self.threshold_upper = params['ChannelWiseThresholdWithNormalize_threshold_upper'.lower()]
        self.mean_std_mode   = params['ChannelWiseThresholdWithNormalize_mean_std_mode'.lower()]
        self.inverse = params.get('ChannelWiseThresholdWithNormalize_inverse'.lower(), False)

    def __call__(self, sample):
        image= sample['image']
        channels = range(image.shape[0]) if self.channels is None else self.channels
        for chn in channels:
            v0 = self.threshold_lower[chn]
            v1 = self.threshold_upper[chn]
            if(self.mean_std_mode == True):
                mask = np.ones_like(image[chn])
                if(v0 is not None):
                    mask = mask * np.asarray(image[chn] > v0)
                if(v1 is not None):
                    mask = mask * np.asarray(image[chn] < v1)
                pixels   = image[chn][mask > 0]
                chn_mean = pixels.mean()
                chn_std  = pixels.std()
                chn_norm = (image[chn] - chn_mean)/chn_std
                chn_random = np.random.normal(0, 1, size = chn_norm.shape)
                chn_norm[mask == 0] = chn_random[mask == 0]
                image[chn] = chn_norm
            else:
                img_chn = image[chn]
                if(v0 is not None):
                    img_chn[img_chn < v0] = v0
                    min_value = v0 
                else:
                    min_value = img_chn.min()
                if(v1 is not None):
                    img_chn[img_chn > v1] = v1 
                    max_value = img_chn.max() 
                else:
                    max_value = img_chn.max() 
                img_chn = (img_chn - min_value) / (max_value - min_value)
                image[chn] = img_chn
        sample['image'] = image
        return sample


# -*- coding: utf-8 -*-
from __future__ import print_function, division
import logging
import numpy as np
import torch
from pymic.loss.seg.util import get_soft_label
from pymic.loss.seg.util import reshape_prediction_and_ground_truth
from pymic.loss.seg.util import get_classwise_dice
from pymic.loss.seg.mumford_shah import MumfordShahLoss
from pymic.net_run.weak_sup import WSLSegAgent
from pymic.util.ramps import get_rampup_ratio

class WSLMumfordShah(WSLSegAgent):
    """
    Weakly supervised learning with Mumford Shah Loss.

    * Reference: Boah Kim and Jong Chul Ye: Mumford–Shah Loss Functional 
      for Image Segmentation With Deep Learning. 
      `IEEE TIP <https://doi.org/10.1109/TIP.2019.2941265>`_, 2019.
         
    :param config: (dict) A dictionary containing the configuration.
    :param stage: (str) One of the stage in `train` (default), `inference` or `test`. 

    .. note::

        In the configuration dictionary, in addition to the four sections (`dataset`,
        `network`, `training` and `inference`) used in fully supervised learning, an 
        extra section `weakly_supervised_learning` is needed. See :doc:`usage.wsl` for details.
    """
    def __init__(self, config, stage = 'train'):
        super(WSLMumfordShah, self).__init__(config, stage)

    def training(self):
        class_num   = self.config['network']['class_num']
        iter_valid  = self.config['training']['iter_valid']
        wsl_cfg     = self.config['weakly_supervised_learning']
        iter_max     = self.config['training']['iter_max']
        rampup_start = wsl_cfg.get('rampup_start', 0)
        rampup_end   = wsl_cfg.get('rampup_end', iter_max)
        train_loss  = 0
        train_loss_sup = 0
        train_loss_reg = 0
        train_dice_list = []

        reg_loss_calculator = MumfordShahLoss(wsl_cfg)
        self.net.train()
        for it in range(iter_valid):
            try:
                data = next(self.trainIter)
            except StopIteration:
                self.trainIter = iter(self.train_loader)
                data = next(self.trainIter)
            
            # get the inputs
            inputs = self.convert_tensor_type(data['image'])
            y      = self.convert_tensor_type(data['label_prob'])  
                         
            inputs, y = inputs.to(self.device), y.to(self.device)
            
            # zero the parameter gradients
            self.optimizer.zero_grad()
                
            # forward + backward + optimize
            outputs = self.net(inputs)
            loss_sup = self.get_loss_value(data, outputs, y)
            loss_dict = {"prediction":outputs, 'image':inputs}
            loss_reg = reg_loss_calculator(loss_dict)
            
            rampup_ratio = get_rampup_ratio(self.glob_it, rampup_start, rampup_end, "sigmoid")
            regular_w = wsl_cfg.get('regularize_w', 0.1) * rampup_ratio
            loss = loss_sup + regular_w*loss_reg
            # if (self.config['training']['use'])
            loss.backward()
            self.optimizer.step()

            train_loss = train_loss + loss.item()
            train_loss_sup = train_loss_sup + loss_sup.item()
            train_loss_reg = train_loss_reg + loss_reg.item() 
            # get dice evaluation for each class in annotated images
            if(isinstance(outputs, tuple) or isinstance(outputs, list)):
                outputs = outputs[0] 
            p_argmax = torch.argmax(outputs, dim = 1, keepdim = True)
            p_soft   = get_soft_label(p_argmax, class_num, self.tensor_type)
            p_soft, y = reshape_prediction_and_ground_truth(p_soft, y) 
            dice_list   = get_classwise_dice(p_soft, y)
            train_dice_list.append(dice_list.cpu().numpy())
        train_avg_loss = train_loss / iter_valid
        train_avg_loss_sup = train_loss_sup / iter_valid
        train_avg_loss_reg = train_loss_reg / iter_valid
        train_cls_dice = np.asarray(train_dice_list).mean(axis = 0)
        train_avg_dice = train_cls_dice[1:].mean()

        train_scalers = {'loss': train_avg_loss, 'loss_sup':train_avg_loss_sup,
            'loss_reg':train_avg_loss_reg, 'regular_w':regular_w,
            'avg_fg_dice':train_avg_dice,     'class_dice': train_cls_dice}
        return train_scalers
        


# -*- coding: utf-8 -*-
from __future__ import print_function, division
import copy 
import json
import math
import random
import numpy as np
from pymic.transform.abstract_transform import AbstractTransform
from pymic.util.image_process import *
try:  # SciPy >= 0.19
    from scipy.special import comb
except ImportError:
    from scipy.misc import comb

def bernstein_poly(i, n, t):
    """
     The Bernstein polynomial of n, i as a function of t
    """

    return comb(n, i) * ( t**(n-i) ) * (1 - t)**i

def bezier_curve(points, nTimes=1000):
    """
       Given a set of control points, return the
       bezier curve defined by the control points.
       Control points should be a list of lists, or list of tuples
       such as [ [1,1], 
                 [2,3], 
                 [4,5], ..[Xn, Yn] ]
        nTimes is the number of time steps, defaults to 1000
        See http://processingjs.nihongoresources.com/bezierinfo/
    """

    nPoints = len(points)
    xPoints = np.array([p[0] for p in points])
    yPoints = np.array([p[1] for p in points])

    t = np.linspace(0.0, 1.0, nTimes)

    polynomial_array = np.array([ bernstein_poly(i, nPoints-1, t) for i in range(0, nPoints)   ])
    
    xvals = np.dot(xPoints, polynomial_array)
    yvals = np.dot(yPoints, polynomial_array)

    return xvals, yvals

class GammaCorrection(AbstractTransform):
    """
    Apply random gamma correction to given channels.

    The arguments should be written in the `params` dictionary, and it has the
    following fields:

    :param `GammaCorrection_channels`: (list) A list of int for specifying the channels.
    :param `GammaCorrection_gamma_min`: (float) The minimal gamma value.
    :param `GammaCorrection_gamma_max`: (float) The maximal gamma value.
    :param `GammaCorrection_probability`: (optional, float) 
        The probability of applying GammaCorrection. Default is 0.5.
    :param `GammaCorrection_inverse`: (optional, bool) 
        Is inverse transform needed for inference. Default is `False`.
    """
    def __init__(self, params):
        super(GammaCorrection, self).__init__(params)
        self.channels =  params['GammaCorrection_channels'.lower()]
        self.gamma_min = params['GammaCorrection_gamma_min'.lower()]
        self.gamma_max = params['GammaCorrection_gamma_max'.lower()]
        self.prob      = params.get('GammaCorrection_probability'.lower(), 0.5)
        self.inverse   = params.get('GammaCorrection_inverse'.lower(), False)
    
    def __call__(self, sample):
        if(np.random.uniform() > self.prob):
            return sample
        image= sample['image']
        for chn in self.channels:
            gamma_c = random.random() * (self.gamma_max - self.gamma_min) + self.gamma_min
            img_c = image[chn]
            v_min = img_c.min()
            v_max = img_c.max()
            img_c = (img_c - v_min)/(v_max - v_min)
            img_c = np.power(img_c, gamma_c)*(v_max - v_min) + v_min
            image[chn] = img_c

        sample['image'] = image
        return sample

class GaussianNoise(AbstractTransform):
    """
    Add Gaussian Noise to given channels.

    The arguments should be written in the `params` dictionary, and it has the
    following fields:

    :param `GaussianNoise_channels`: (list) A list of int for specifying the channels.
    :param `GaussianNoise_mean`: (float) The mean value of noise.
    :param `GaussianNoise_std`: (float) The std of noise.
    :param `GaussianNoise_probability`: (optional, float) 
        The probability of applying GaussianNoise. Default is 0.5.
    :param `GaussianNoise_inverse`: (optional, bool) 
        Is inverse transform needed for inference. Default is `False`.
    """
    def __init__(self, params):
        super(GaussianNoise, self).__init__(params)
        self.channels = params['GaussianNoise_channels'.lower()]
        self.mean     = params['GaussianNoise_mean'.lower()]
        self.std      = params['GaussianNoise_std'.lower()]
        self.prob     = params.get('GaussianNoise_probability'.lower(), 0.5)
        self.inverse  = params.get('GaussianNoise_inverse'.lower(), False)
    
    def __call__(self, sample):
        if(np.random.uniform() > self.prob):
            return sample
        image= sample['image']
        for chn in self.channels:
            img_c = image[chn]
            noise = np.random.normal(self.mean, self.std, img_c.shape)
            image[chn] = img_c + noise

        sample['image'] = image
        return sample

class GrayscaleToRGB(AbstractTransform):
    """
    Convert gray scale images to RGB by copying channels. 
    """
    def __init__(self, params):
        super(GrayscaleToRGB, self).__init__(params)
        self.inverse = params.get('GrayscaleToRGB_inverse'.lower(), False)
    
    def __call__(self, sample):
        image= sample['image']
        assert(image.shape[0] == 1 or image.shape[0] == 3)
        if(image.shape[0] == 1):
            sample['image'] = np.concatenate([image, image, image])
        return sample
    
class NonLinearTransform(AbstractTransform):
    def __init__(self, params):
        super(NonLinearTransform, self).__init__(params)
        self.inverse  = params.get('NonLinearTransform_inverse'.lower(), False)
        self.prob     = params.get('NonLinearTransform_probability'.lower(), 0.5)
    
    def __call__(self, sample):
        if(random.random() >  self.prob):
            return sample

        image= sample['image'] 
        points = [[0, 0], [random.random(), random.random()], [random.random(), random.random()], [1, 1]]
        xvals, yvals = bezier_curve(points, nTimes=100000)
        if random.random() < 0.5: # Half change to get flip
            xvals = np.sort(xvals)
        else:
            xvals, yvals = np.sort(xvals), np.sort(yvals)
        image = np.interp(image, xvals, yvals)
        sample['image']  = image 
        return sample

class LocalShuffling(AbstractTransform):
    """
    local pixel shuffling of an input image, used for self-supervised learning 
    """
    def __init__(self, params):
        super(LocalShuffling, self).__init__(params)
        self.inverse  = params.get('LocalShuffling_inverse'.lower(), False)
        self.prob     = params.get('LocalShuffling_probability'.lower(), 0.5)
        self.block_range = params.get('LocalShuffling_block_range'.lower(), (5000, 10000))
        self.block_size_min = params.get('LocalShuffling_block_size_min'.lower(), None)
        self.block_size_max = params.get('LocalShuffling_block_size_max'.lower(), None)

    def __call__(self, sample):
        if(random.random() >  self.prob):
            return sample

        image= sample['image']       
        img_shape = image.shape
        img_dim = len(img_shape) - 1
        assert(img_dim == 2 or img_dim == 3)
        img_out = copy.deepcopy(image)
        if(self.block_size_min is None):
            block_size_min = [2] * img_dim
        elif(isinstance(self.block_size_min, int)):
            block_size_min = [self.block_size_min] * img_dim
        else:
            assert(len(self.block_size_min) == img_dim)
            block_size_min = self.block_size_min

        if(self.block_size_max is None):
            block_size_max = [img_shape[1+i]//10 for i in range(img_dim)]
        elif(isinstance(self.block_size_min, int)):
            block_size_max = [self.block_size_max] * img_dim
        else:
            assert(len(self.block_size_max) == img_dim)
            block_size_max = self.block_size_max
        block_num = random.randint(self.block_range[0], self.block_range[1])

        for n in range(block_num):
            block_size = [random.randint(block_size_min[i], block_size_max[i]) \
                for i in range(img_dim)]    
            coord_min = [random.randint(0, img_shape[1+i] - block_size[i]) \
                for i in range(img_dim)]
            if(img_dim == 2):
                window = image[:, coord_min[0]:coord_min[0] + block_size[0], 
                                  coord_min[1]:coord_min[1] + block_size[1]]
                n_pixels = block_size[0] * block_size[1]
            else:
                window = image[:, coord_min[0]:coord_min[0] + block_size[0], 
                                  coord_min[1]:coord_min[1] + block_size[1],
                                  coord_min[2]:coord_min[2] + block_size[2]]
                n_pixels = block_size[0] * block_size[1] * block_size[2]
            window = np.reshape(window, [-1, n_pixels])
            np.random.shuffle(np.transpose(window))
            window = np.transpose(window)
            if(img_dim == 2):
                window = np.reshape(window, [-1, block_size[0], block_size[1]])
                img_out[:, coord_min[0]:coord_min[0] + block_size[0], 
                           coord_min[1]:coord_min[1] + block_size[1]] = window
            else:
                window = np.reshape(window, [-1, block_size[0], block_size[1], block_size[2]])
                img_out[:, coord_min[0]:coord_min[0] + block_size[0], 
                           coord_min[1]:coord_min[1] + block_size[1],
                           coord_min[2]:coord_min[2] + block_size[2]] = window
        sample['image'] = img_out
        return sample

class InPainting(AbstractTransform):
    """
    In-painting of an input image, used for self-supervised learning 
    """
    def __init__(self, params):
        super(InPainting, self).__init__(params)
        self.inverse  = params.get('InPainting_inverse'.lower(), False)
        self.prob     = params.get('InPainting_probability'.lower(), 0.5)
        self.block_range = params.get('InPainting_block_range'.lower(), (1, 6))
        self.block_size_min = params.get('InPainting_block_size_min'.lower(), None)
        self.block_size_max = params.get('InPainting_block_size_max'.lower(), None)

    def __call__(self, sample):
        if(random.random() >  self.prob):
            return sample

        image= sample['image']       
        img_shape = image.shape
        img_dim = len(img_shape) - 1
        assert(img_dim == 2 or img_dim == 3)

        if(self.block_size_min is None):
            block_size_min = [img_shape[1+i]//6 for i in range(img_dim)]
        elif(isinstance(self.block_size_min, int)):
            block_size_min = [self.block_size_min] * img_dim
        else:
            assert(len(self.block_size_min) == img_dim)
            block_size_min = self.block_size_min

        if(self.block_size_max is None):
            block_size_max = [img_shape[1+i]//3 for i in range(img_dim)]
        elif(isinstance(self.block_size_min, int)):
            block_size_max = [self.block_size_max] * img_dim
        else:
            assert(len(self.block_size_max) == img_dim)
            block_size_max = self.block_size_max
        block_num = random.randint(self.block_range[0], self.block_range[1])

        for n in range(block_num):
            block_size = [random.randint(block_size_min[i], block_size_max[i]) \
                for i in range(img_dim)]    
            coord_min = [random.randint(3, img_shape[1+i] - block_size[i] - 3) \
                for i in range(img_dim)]
            if(img_dim == 2):
                random_block = np.random.rand(img_shape[0], block_size[0], block_size[1])
                image[:, coord_min[0]:coord_min[0] + block_size[0], 
                         coord_min[1]:coord_min[1] + block_size[1]] = random_block
            else:
                random_block = np.random.rand(img_shape[0], block_size[0], 
                                              block_size[1], block_size[2])
                image[:, coord_min[0]:coord_min[0] + block_size[0], 
                         coord_min[1]:coord_min[1] + block_size[1],
                         coord_min[2]:coord_min[2] + block_size[2]] = random_block
        sample['image'] = image
        return sample

class OutPainting(AbstractTransform):
    """
    Out-painting of an input image, used for self-supervised learning 
    """
    def __init__(self, params):
        super(OutPainting, self).__init__(params)
        self.inverse  = params.get('OutPainting_inverse'.lower(), False)
        self.prob     = params.get('OutPainting_probability'.lower(), 0.5)
        self.block_range = params.get('OutPainting_block_range'.lower(), (1, 6))
        self.block_size_min = params.get('OutPainting_block_size_min'.lower(), None)
        self.block_size_max = params.get('OutPainting_block_size_max'.lower(), None)

    def __call__(self, sample):
        if(random.random() >  self.prob):
            return sample

        image= sample['image']       
        img_shape = image.shape
        img_dim = len(img_shape) - 1
        assert(img_dim == 2 or img_dim == 3)
        img_out = np.random.rand(*img_shape)

        if(self.block_size_min is None):
            block_size_min = [img_shape[1+i] - 4 * img_shape[1+i]//7 for i in range(img_dim)]
        elif(isinstance(self.block_size_min, int)):
            block_size_min = [self.block_size_min] * img_dim
        else:
            assert(len(self.block_size_min) == img_dim)
            block_size_min = self.block_size_min

        if(self.block_size_max is None):
            block_size_max = [img_shape[1+i] - 3 * img_shape[1+i]//7 for i in range(img_dim)]
        elif(isinstance(self.block_size_min, int)):
            block_size_max = [self.block_size_max] * img_dim
        else:
            assert(len(self.block_size_max) == img_dim)
            block_size_max = self.block_size_max
        block_num = random.randint(self.block_range[0], self.block_range[1])

        for n in range(block_num):
            block_size = [random.randint(block_size_min[i], block_size_max[i]) \
                for i in range(img_dim)]    
            coord_min = [random.randint(3, img_shape[1+i] - block_size[i] - 3) \
                for i in range(img_dim)]
            if(img_dim == 2):
                img_out[:, coord_min[0]:coord_min[0] + block_size[0], 
                           coord_min[1]:coord_min[1] + block_size[1]] = \
                    image[:, coord_min[0]:coord_min[0] + block_size[0], 
                             coord_min[1]:coord_min[1] + block_size[1]]
            else:
                img_out[:, coord_min[0]:coord_min[0] + block_size[0], 
                           coord_min[1]:coord_min[1] + block_size[1],
                           coord_min[2]:coord_min[2] + block_size[2]] = \
                    image[:, coord_min[0]:coord_min[0] + block_size[0], 
                             coord_min[1]:coord_min[1] + block_size[1],
                             coord_min[2]:coord_min[2] + block_size[2]] 
        sample['image'] = img_out
        return sample

class InOutPainting(AbstractTransform):
    """
    Apply in-painting or out-patining randomly. They are mutually exclusive.
    """
    def __init__(self, params):
        super(InOutPainting, self).__init__(params)
        self.inverse  = params.get('InOutPainting_inverse'.lower(), False)
        self.prob     = params.get('InOutPainting_probability'.lower(), 0.5)
        self.in_prob  = params.get('InPainting_probability'.lower(), 0.5)
        params['InPainting_probability']  = 1.0
        params['outPainting_probability'] = 1.0
        self.inpaint  = InPainting(params)
        self.outpaint = OutPainting(params)

    def __call__(self, sample):
        if(random.random() >  self.prob):
            return sample
        if(random.random() < self.in_prob):
            sample = self.inpaint(sample)
        else:
            sample = self.outpaint(sample)
        return sample


# -*- coding: utf-8 -*-
from __future__ import print_function, division

import torch
import json
import math
import random
import numpy as np
from scipy import ndimage
from pymic.transform.abstract_transform import AbstractTransform
from pymic.util.image_process import *

class ReduceLabelDim(AbstractTransform):
    """
    Remove the first dimension of label tensor.
    """
    def __init__(self, params):
        super(ReduceLabelDim, self).__init__(params)
        self.inverse = params.get('ReduceLabelDim_inverse'.lower(), False)
    
    def __call__(self, sample):
        label = sample['label']
        label_converted = label[0]
        sample['label'] = label_converted
        return sample

class LabelConvert(AbstractTransform):
    """ 
    Convert the label based on a source list and target list. 
    
    The arguments should be written in the `params` dictionary, and it has the
    following fields:

    :param `LabelConvert_source_list`: (list) A list of labels to be converted.
    :param `LabelConvert_target_list`: (list) The target label list.
    :param `LabelConvert_inverse`: (optional, bool) 
        Is inverse transform needed for inference. Default is `False`.
    """
    def __init__(self, params):
        super(LabelConvert, self).__init__(params)
        self.source_list = params['LabelConvert_source_list'.lower()]
        self.target_list = params['LabelConvert_target_list'.lower()]
        self.inverse = params.get('LabelConvert_inverse'.lower(), False)
        assert(len(self.source_list) == len(self.target_list))
    
    def __call__(self, sample):
        label = sample['label']
        label_converted = convert_label(label, self.source_list, self.target_list)
        sample['label'] = label_converted
        return sample

class LabelConvertNonzero(AbstractTransform):
    """ 
    Convert label into binary, i.e., setting nonzero labels as 1.
    """
    def __init__(self, params):
        super(LabelConvertNonzero, self).__init__(params)
        self.inverse = params.get('LabelConvertNonzero_inverse'.lower(), False)
    
    def __call__(self, sample):
        label = sample['label']
        label_converted = np.asarray(label > 0, np.uint8)
        sample['label'] = label_converted
        return sample

class LabelToProbability(AbstractTransform):
    """
    Convert one-channel label map to one-hot multi-channel probability map.
    
    The arguments should be written in the `params` dictionary, and it has the
    following fields:

    :param `LabelToProbability_class_num`: (int) The class number in the label map.
    :param `LabelToProbability_inverse`: (optional, bool) 
        Is inverse transform needed for inference. Default is `False`.
    """
    def __init__(self, params): 
        super(LabelToProbability, self).__init__(params)
        self.class_num = params['LabelToProbability_class_num'.lower()]
        self.inverse   = params.get('LabelToProbability_inverse'.lower(), False)

    def __call__(self, sample):
        if(self.task == 'segmentation'):
            label = sample['label'][0] # sample['label'] is (1, h, w)
            label_prob = np.zeros((self.class_num, *label.shape), dtype = np.float32)
            for i in range(self.class_num):
                label_prob[i] = label == i*np.ones_like(label)
            sample['label_prob'] = label_prob
        elif(self.task == 'classification'):
            label_idx = sample['label']
            label_prob = np.zeros((self.class_num,), np.float32)
            label_prob[label_idx] = 1.0
            sample['label_prob'] = label_prob 
        return sample

class LabelSmooth(AbstractTransform):
    """
    Apply label smoothing to one-hot labels.
    
    The arguments should be written in the `params` dictionary, and it has the
    following fields:

    :param `LabelSmooth_alpha`: (float) Alpha value for label smoothing.
    :param `LabelSmooth_inverse`: (optional, bool) 
        Is inverse transform needed for inference. Default is `False`.
    """
    def __init__(self, params): 
        super(LabelSmooth, self).__init__(params)
        self.alpha = params['LabelSmooth_alpha'.lower()]
        self.inverse   = params.get('LabelSmooth_inverse'.lower(), False)

    def __call__(self, sample):
        label_prob = sample['label_prob']
        K = list(label_prob.shape)[1]
        sample['label_prob'] = label_prob * (1.0 - self.alpha) + self.alpha / K
        return sample

class PartialLabelToProbability(AbstractTransform):
    """
    Convert one-channel partial label map to one-hot multi-channel probability map.
    This is used for segmentation tasks only. In the input label map, 0 represents the
    background class, 1 to C-1 represent the foreground classes, and C represents 
    unlabeled pixels. In the output dictionary, `label_prob` is the one-hot probability 
    map, and `pixel_weight` represents a weighting map, where the weight for a pixel
    is 0 if the label is unkown. 

    The arguments should be written in the `params` dictionary, and it has the
    following fields:

    :param `PartialLabelToProbability_class_num`: (int) The class number for the 
        segmentation task.  
    :param `PartialLabelToProbability_inverse`: (optional, bool) 
        Is inverse transform needed for inference. Default is `False`.
    """
    def __init__(self, params): 
        """
        class_num (int): the class number in the label map
        """
        super(PartialLabelToProbability, self).__init__(params)
        self.class_num = params['PartialLabelToProbability_class_num'.lower()]
        self.inverse   = params.get('PartialLabelToProbability_inverse'.lower(), False)
    
    def __call__(self, sample):
        label = sample['label'][0]
        assert(label.max() <= self.class_num)
        label_prob = np.zeros((self.class_num, *label.shape), dtype = np.float32)
        for i in range(self.class_num):
            label_prob[i] = label == i*np.ones_like(label)
        sample['label_prob'] = label_prob
        sample['pixel_weight'] = 1.0 - np.asarray([label == self.class_num], np.float32)
        return sample







# -*- coding: utf-8 -*-
from __future__ import print_function, division

class AbstractTransform(object):
    """
    The abstract class for Transform.
    """
    def __init__(self, params):
        self.task = params['Task'.lower()]

    def __call__(self, sample):
        """
        Forward pass of the transform. 

        :arg sample: (dict) A dictionary for the input sample obtained by dataloader.
        """
        return sample

    def inverse_transform_for_prediction(self, sample):
        """
        Inverse transform for the sample dictionary.
        Especially, it will update sample['predict'] obtained by a network's
        prediction based on the inverse transform. This function is only useful for spatial transforms.
        """
        raise(ValueError("not implemented"))



# -*- coding: utf-8 -*-
from __future__ import print_function, division
import torch
import numpy as np 

def keyword_match(a,b):
    """
    Test if two string are the same when converted to lower case. 
    """
    return a.lower() == b.lower()

def tensor_shape_match(a,b):
    """
    Test if two tensors have the same shape"""
    shape_a = list(a.shape)
    shape_b = list(b.shape)
    len_a  = len(shape_a)
    len_b  = len(shape_b)
    if(len_a != len_b):
        return False 
    elif(len_a == 0):
        return True 
    else:
        for i in range(len_a):
            if(shape_a[i] != shape_b[i]):
                return False 
        return True 



def get_one_hot_seg(label, class_num):
    """
    Convert a segmentation label to one-hot.

    :param label: A tensor with a shape of [N, 1, D, H, W] or [N, 1, H, W]
    :param class_num: Class number. 
    
    :return: a one-hot tensor with a shape of [N, C, D, H, W] or [N, C, H, W].
    """
    size = list(label.size())
    if(size[1] != 1):
        raise ValueError("The channel should be 1, \
            rather than {0:} before one-hot encoding".format(size[1]))
    label = label.view(-1)
    ones  = torch.sparse.torch.eye(class_num).to(label.device)
    one_hot = ones.index_select(0, label)
    size.append(class_num)
    one_hot = one_hot.view(*size)
    one_hot = torch.transpose(one_hot, 1, -1)
    one_hot = torch.squeeze(one_hot, -1)
    return one_hot

def mixup(inputs, labels):
    """Shuffle a minibatch and do linear interpolation between images and labels.
    Both classification and segmentation labels are supported. The targets should
    be one-hot labels.
    
    :param inputs: a tensor of input images with size N X C0 x H x W.
    :param labels: a tensor of one-hot labels. The shape is N X C for classification
        tasks, and N X C X H X W for segmentation tasks. 
    """
    input_shape = list(inputs.shape)
    label_shape = list(labels.shape)
    img_dim     = len(input_shape) - 2
    N = input_shape[0] # batch size
    C = label_shape[1] # class number
    rp1 = torch.randperm(N)
    inputs1 = inputs[rp1]
    labels1 = labels[rp1]
    
    rp2 = torch.randperm(N)
    inputs2 = inputs[rp2]
    labels2 = labels[rp2]

    a = np.random.beta(1, 1, [N, 1])
    if(img_dim == 2):
        b = np.tile(a[..., None, None], [1] + input_shape[1:])
    elif(img_dim == 3):
        b = np.tile(a[..., None, None, None], [1] + input_shape[1:])
    else:
        raise ValueError("MixUp only supports 2D and 3D images, but the " +
            "input image has {0:} dimensions".format(img_dim))

    inputs1 = inputs1 * torch.from_numpy(b).float()
    inputs2 = inputs2 * torch.from_numpy(1 - b).float()
    inputs_mix = inputs1 + inputs2

    if(len(label_shape) == 2): # for classification tasks
        c = np.tile(a, [1, C])
    elif(img_dim == 2):        # for 2D segmentation tasks
        c = np.tile(a[..., None, None], [1] + label_shape[1:])
    else:                      # for 3D segmentation tasks
        c = np.tile(a[..., None, None, None], [1] + label_shape[1:])
    
    labels1 = labels1 * torch.from_numpy(c).float()
    labels2 = labels2 * torch.from_numpy(1 - c).float()
    labels_mix = labels1 + labels2

    return inputs_mix, labels_mix



# -*- coding: utf-8 -*-
from __future__ import print_function, division

import torch
import torch.nn as nn

class ConvolutionLayer(nn.Module):
    """
    A compose layer with the following components:
    convolution -> (batch_norm / layer_norm / group_norm / instance_norm) -> (activation) -> (dropout)
    Batch norm and activation are optional.

    :param in_channels: (int) The input channel number.
    :param out_channels: (int) The output channel number. 
    :param kernel_size: The size of convolution kernel. It can be either a single 
        int or a tupe of two or three ints. 
    :param dim: (int) The dimention of convolution (2 or 3).
    :param stride: (int) The stride of convolution. 
    :param padding: (int) Padding size. 
    :param dilation: (int) Dilation rate.
    :param conv_group: (int) The groupt number of convolution. 
    :param bias: (bool) Add bias or not for convolution. 
    :param norm_type: (str or None) Normalization type, can be `batch_norm`, 'group_norm'.
    :param norm_group: (int) The number of group for group normalization.
    :param acti_func: (str or None) Activation funtion. 
    """
    def __init__(self, in_channels, out_channels, kernel_size, dim = 3,
            stride = 1, padding = 0, dilation = 1, conv_group = 1, bias = True, 
            norm_type = 'batch_norm', norm_group = 1, acti_func = None):
        super(ConvolutionLayer, self).__init__()
        self.n_in_chns  = in_channels
        self.n_out_chns = out_channels
        self.norm_type  = norm_type
        self.norm_group = norm_group
        self.acti_func  = acti_func

        assert(dim == 2 or dim == 3)
        if(dim == 2):
            self.conv = nn.Conv2d(in_channels, out_channels,
                kernel_size, stride, padding, dilation, conv_group, bias)
            if(self.norm_type == 'batch_norm'):
                self.bn = nn.BatchNorm2d(out_channels)
            elif(self.norm_type == 'group_norm'):
                self.bn = nn.GroupNorm(self.norm_group, out_channels)
            elif(self.norm_type is not None):
                raise ValueError("unsupported normalization method {0:}".format(norm_type))
        else:        
            self.conv = nn.Conv3d(in_channels, out_channels,
                kernel_size, stride, padding, dilation, conv_group, bias)
            if(self.norm_type == 'batch_norm'):
                self.bn = nn.BatchNorm3d(out_channels)
            elif(self.norm_type == 'group_norm'):
                self.bn = nn.GroupNorm(self.norm_group, out_channels)
            elif(self.norm_type is not None):
                raise ValueError("unsupported normalization method {0:}".format(norm_type))

    def forward(self, x):
        f = self.conv(x)
        if(self.norm_type is not None):
            f = self.bn(f)
        if(self.acti_func is not None):
            f = self.acti_func(f)
        return f

class DepthSeperableConvolutionLayer(nn.Module):
    """
    Depth seperable convolution with the following components:
    1x1 conv -> group conv -> (batch_norm / layer_norm / group_norm / instance_norm) -> (activation) -> (dropout)
    Batch norm and activation are optional.

    :param in_channels: (int) The input channel number.
    :param out_channels: (int) The output channel number. 
    :param kernel_size: The size of convolution kernel. It can be either a single 
        int or a tupe of two or three ints. 
    :param dim: (int) The dimention of convolution (2 or 3).
    :param stride: (int) The stride of convolution. 
    :param padding: (int) Padding size. 
    :param dilation: (int) Dilation rate.
    :param conv_group: (int) The groupt number of convolution. 
    :param bias: (bool) Add bias or not for convolution. 
    :param norm_type: (str or None) Normalization type, can be `batch_norm`, 'group_norm'.
    :param norm_group: (int) The number of group for group normalization.
    :param acti_func: (str or None) Activation funtion. 
    """
    def __init__(self, in_channels, out_channels, kernel_size, dim = 3,
            stride = 1, padding = 0, dilation =1, conv_group = 1, bias = True, 
            norm_type = 'batch_norm', norm_group = 1, acti_func = None):
        super(DepthSeperableConvolutionLayer, self).__init__()
        self.n_in_chns  = in_channels
        self.n_out_chns = out_channels
        self.norm_type  = norm_type
        self.norm_group = norm_group
        self.acti_func  = acti_func

        assert(dim == 2 or dim == 3)
        if(dim == 2):
            self.conv1x1 = nn.Conv2d(in_channels, out_channels,
                kernel_size = 1, stride = stride, padding = 0, dilation = dilation, groups = conv_group, bias = bias)
            self.conv = nn.Conv2d(out_channels, out_channels,
                kernel_size, stride, padding, dilation, groups = out_channels, bias = bias)
            if(self.norm_type == 'batch_norm'):
                self.bn = nn.BatchNorm2d(out_channels)
            elif(self.norm_type == 'group_norm'):
                self.bn = nn.GroupNorm(self.norm_group, out_channels)
            elif(self.norm_type is not None):
                raise ValueError("unsupported normalization method {0:}".format(norm_type))
        else:     
            self.conv1x1 = nn.Conv3d(in_channels, out_channels,
                kernel_size = 1, stride = stride, padding = 0, dilation = dilation, groups = conv_group, bias = bias)   
            self.conv = nn.Conv3d(out_channels, out_channels,
                kernel_size, stride, padding, dilation, groups = out_channels, bias = bias)
            if(self.norm_type == 'batch_norm'):
                self.bn = nn.BatchNorm3d(out_channels)
            elif(self.norm_type == 'group_norm'):
                self.bn = nn.GroupNorm(self.norm_group, out_channels)
            elif(self.norm_type is not None):
                raise ValueError("unsupported normalization method {0:}".format(norm_type))

    def forward(self, x):
        f = self.conv1x1(x)
        f = self.conv(f)
        if(self.norm_type is not None):
            f = self.bn(f)
        if(self.acti_func is not None):
            f = self.acti_func(f)
        return f




# -*- coding: utf-8 -*-
from __future__ import print_function, division

import math
import copy
import torch
import torch.nn as nn
import numpy as np
import torch.nn.functional as F
from pymic.net.net3d.unet3d import ConvBlock, Encoder, Decoder
from pymic.net.net3d.trans3d.MedFormer_v1 import Block
from timm.models.layers import DropPath, to_3tuple, trunc_normal_


# code from nnFormer
class Mlp(nn.Module):
    """ Multilayer perceptron."""

    def __init__(self, in_features, hidden_features=None, out_features=None, act_layer=nn.GELU, drop=0.):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x


def window_partition(x, window_size):
  
    B, S, H, W, C = x.shape
    x = x.view(B, S // window_size, window_size, H // window_size, window_size, W // window_size, window_size, C)
    windows = x.permute(0, 1, 3, 5, 2, 4, 6, 7).contiguous().view(-1, window_size, window_size, window_size, C)
    return windows


def window_reverse(windows, window_size, S, H, W):
   
    B = int(windows.shape[0] / (S * H * W / window_size / window_size / window_size))
    x = windows.view(B, S // window_size, H // window_size, W // window_size, window_size, window_size, window_size, -1)
    x = x.permute(0, 1, 4, 2, 5, 3, 6, 7).contiguous().view(B, S, H, W, -1)
    return x


class WindowAttention(nn.Module):

    def __init__(self, dim, window_size, num_heads, qkv_bias=True, qk_scale=None, attn_drop=0., proj_drop=0.):

        super().__init__()
        self.dim = dim
        self.window_size = window_size  
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = qk_scale or head_dim ** -0.5

        # define a parameter table of relative position bias
        self.relative_position_bias_table = nn.Parameter(
            torch.zeros((2 * window_size[0] - 1) * (2 * window_size[1] - 1) * (2 * window_size[2] - 1),
                        num_heads)) 

        # get pair-wise relative position index for each token inside the window
        coords_s = torch.arange(self.window_size[0])
        coords_h = torch.arange(self.window_size[1])
        coords_w = torch.arange(self.window_size[2])
        coords = torch.stack(torch.meshgrid([coords_s, coords_h, coords_w]))  
        coords_flatten = torch.flatten(coords, 1)  
        relative_coords = coords_flatten[:, :, None] - coords_flatten[:, None, :]  
        relative_coords = relative_coords.permute(1, 2, 0).contiguous() 
        relative_coords[:, :, 0] += self.window_size[0] - 1  # shift to start from 0
        relative_coords[:, :, 1] += self.window_size[1] - 1
        relative_coords[:, :, 2] += self.window_size[2] - 1

        relative_coords[:, :, 0] *= 3 * self.window_size[1] - 1
        relative_coords[:, :, 1] *= 2 * self.window_size[1] - 1

        relative_position_index = relative_coords.sum(-1) 
        self.register_buffer("relative_position_index", relative_position_index)

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)
        trunc_normal_(self.relative_position_bias_table, std=.02)

        self.softmax = nn.Softmax(dim=-1)

    def forward(self, x, mask=None,pos_embed=None):

        B_, N, C = x.shape
        
        qkv = self.qkv(x)
        
        qkv=qkv.reshape(B_, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4).contiguous()
        q, k, v = qkv[0], qkv[1], qkv[2]  # make torchscript happy (cannot use tensor as tuple)
        
        q = q * self.scale
        attn = (q @ k.transpose(-2, -1).contiguous())
        relative_position_bias = self.relative_position_bias_table[self.relative_position_index.view(-1)].view(
            self.window_size[0] * self.window_size[1] * self.window_size[2],
            self.window_size[0] * self.window_size[1] * self.window_size[2], -1)  
        relative_position_bias = relative_position_bias.permute(2, 0, 1).contiguous() 
        attn = attn + relative_position_bias.unsqueeze(0)

        if mask is not None:
            nW = mask.shape[0]
            attn = attn.view(B_ // nW, nW, self.num_heads, N, N) + mask.unsqueeze(1).unsqueeze(0)
            attn = attn.view(-1, self.num_heads, N, N)
            attn = self.softmax(attn)
        else:
            attn = self.softmax(attn)

        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B_, N, C).contiguous()
        if pos_embed is not None:
            x = x+pos_embed
        x = self.proj(x)
        x = self.proj_drop(x)
        return x

class SwinTransformerBlock(nn.Module):
    
    def __init__(self, dim, input_resolution, num_heads, window_size=7, shift_size=0,
                 mlp_ratio=4., qkv_bias=True, qk_scale=None, drop=0., attn_drop=0., drop_path=0.,
                 act_layer=nn.GELU, norm_layer=nn.LayerNorm):
        super().__init__()
        self.dim = dim
        self.input_resolution = input_resolution
        self.num_heads = num_heads
        self.window_size = window_size
        self.shift_size = shift_size
        self.mlp_ratio = mlp_ratio
   
        if min(self.input_resolution) <= self.window_size:
            # if window size is larger than input resolution, we don't partition windows
            self.shift_size = 0
            self.window_size = min(self.input_resolution)

        assert 0 <= self.shift_size < self.window_size, "shift_size must in 0-window_size"

        self.norm1 = norm_layer(dim)
        
        self.attn = WindowAttention(
            dim, window_size=to_3tuple(self.window_size), num_heads=num_heads,
            qkv_bias=qkv_bias, qk_scale=qk_scale, attn_drop=attn_drop, proj_drop=drop)
       
            

        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(in_features=dim, hidden_features=mlp_hidden_dim, act_layer=act_layer, drop=drop)
        
       
    def forward(self, x, mask_matrix):

        B, L, C = x.shape
        S, H, W = self.input_resolution
   
        assert L == S * H * W, "input feature has wrong size"
        
        shortcut = x
        x = self.norm1(x)
        x = x.view(B, S, H, W, C)

        # pad feature maps to multiples of window size
        pad_r = (self.window_size - W % self.window_size) % self.window_size
        pad_b = (self.window_size - H % self.window_size) % self.window_size
        pad_g = (self.window_size - S % self.window_size) % self.window_size

        x = F.pad(x, (0, 0, 0, pad_r, 0, pad_b, 0, pad_g))  
        _, Sp, Hp, Wp, _ = x.shape

        # cyclic shift
        if self.shift_size > 0:
            shifted_x = torch.roll(x, shifts=(-self.shift_size, -self.shift_size,-self.shift_size), dims=(1, 2,3))
            attn_mask = mask_matrix
        else:
            shifted_x = x
            attn_mask = None
       
        # partition windows
        x_windows = window_partition(shifted_x, self.window_size)  # nW*B, window_size, window_size, C
        x_windows = x_windows.view(-1, self.window_size * self.window_size * self.window_size,
                                   C)  

        # W-MSA/SW-MSA
        attn_windows = self.attn(x_windows, mask=attn_mask,pos_embed=None)  

        # merge windows
        attn_windows = attn_windows.view(-1, self.window_size, self.window_size, self.window_size, C)
        shifted_x = window_reverse(attn_windows, self.window_size, Sp, Hp, Wp) 

        # reverse cyclic shift
        if self.shift_size > 0:
            x = torch.roll(shifted_x, shifts=(self.shift_size, self.shift_size, self.shift_size), dims=(1, 2, 3))
        else:
            x = shifted_x

        if pad_r > 0 or pad_b > 0 or pad_g > 0:
            x = x[:, :S, :H, :W, :].contiguous()

        x = x.view(B, S * H * W, C)

        # FFN
        x = shortcut + self.drop_path(x)
        x = x + self.drop_path(self.mlp(self.norm2(x)))

        return x

class BasicLayer(nn.Module):
   
    def __init__(self,
                 dim,
                 input_resolution,
                 depth,
                 num_heads,
                 window_size=7,
                 mlp_ratio=4.,
                 qkv_bias=True,
                 qk_scale=None,
                 drop=0.,
                 attn_drop=0.,
                 drop_path=0.,
                 norm_layer=nn.LayerNorm,
                 downsample=True
                 ):
        super().__init__()
        self.window_size = window_size
        self.shift_size = window_size // 2
        self.depth = depth
        # build blocks
        
        self.blocks = nn.ModuleList([
            SwinTransformerBlock(
                dim=dim,
                input_resolution=input_resolution,
                num_heads=num_heads,
                window_size=window_size,
                shift_size=0 if (i % 2 == 0) else window_size // 2,
                mlp_ratio=mlp_ratio,
                qkv_bias=qkv_bias,
                qk_scale=qk_scale,
                drop=drop,
                attn_drop=attn_drop,
                drop_path=drop_path[i] if isinstance(drop_path, list) else drop_path, norm_layer=norm_layer)
            for i in range(depth)])

        # patch merging layer
        if downsample is not None:
            self.downsample = downsample(dim=dim, norm_layer=norm_layer)
        else:
            self.downsample = None

    def forward(self, x, S, H, W):
      

        # calculate attention mask for SW-MSA
        Sp = int(np.ceil(S / self.window_size)) * self.window_size
        Hp = int(np.ceil(H / self.window_size)) * self.window_size
        Wp = int(np.ceil(W / self.window_size)) * self.window_size
        img_mask = torch.zeros((1, Sp, Hp, Wp, 1), device=x.device)  # 1 Hp Wp 1
        s_slices = (slice(0, -self.window_size),
                    slice(-self.window_size, -self.shift_size),
                    slice(-self.shift_size, None))
        h_slices = (slice(0, -self.window_size),
                    slice(-self.window_size, -self.shift_size),
                    slice(-self.shift_size, None))
        w_slices = (slice(0, -self.window_size),
                    slice(-self.window_size, -self.shift_size),
                    slice(-self.shift_size, None))
        cnt = 0
        for s in s_slices:
            for h in h_slices:
                for w in w_slices:
                    img_mask[:, s, h, w, :] = cnt
                    cnt += 1

        mask_windows = window_partition(img_mask, self.window_size)  
        mask_windows = mask_windows.view(-1,
                                         self.window_size * self.window_size * self.window_size)  
        attn_mask = mask_windows.unsqueeze(1) - mask_windows.unsqueeze(2)
        attn_mask = attn_mask.masked_fill(attn_mask != 0, float(-100.0)).masked_fill(attn_mask == 0, float(0.0))
        for blk in self.blocks:
          
            x = blk(x, attn_mask)
        if self.downsample is not None:
            x_down = self.downsample(x, S, H, W)
            Ws, Wh, Ww = (S + 1) // 2, (H + 1) // 2, (W + 1) // 2
            return x, S, H, W, x_down, Ws, Wh, Ww
        else:
            return x, S, H, W, x, S, H, W


class AttUpBlock(nn.Module):
    """
    3D upsampling followed by ConvBlock
    
    :param in_channels1: (int) Channel number of high-level features.
    :param in_channels2: (int) Channel number of low-level features.
    :param out_channels: (int) Output channel number.
    :param dropout_p: (int) Dropout probability.
    :param trilinear: (bool) Use trilinear for up-sampling (by default).
        If False, deconvolution is used for up-sampling. 
    """
    def __init__(self, in_channels1, in_channels2, out_channels, dropout_p,
                 trilinear=True, with_att = False, att_params = None):
        super(AttUpBlock, self).__init__()
        self.trilinear = trilinear
        self.with_att  = with_att
        if trilinear:
            self.conv1x1 = nn.Conv3d(in_channels1, in_channels2, kernel_size = 1)
            self.up = nn.Upsample(scale_factor=2, mode='trilinear', align_corners=True)
        else:
            self.up = nn.ConvTranspose3d(in_channels1, in_channels2, kernel_size=2, stride=2)
        self.conv = ConvBlock(in_channels2 * 2, out_channels, dropout_p)
        if(self.with_att):
            input_resolution = att_params['input_resolution']
            depth            = att_params['depth']
            num_heads        = att_params['num_heads']
            self.attn  = BasicLayer(out_channels, input_resolution, depth, num_heads, downsample=None)

    def forward(self, x1, x2):
        if self.trilinear:
            x1 = self.conv1x1(x1)
        x1 = self.up(x1)
        x = torch.cat([x2, x1], dim=1)
        x =  self.conv(x) 
        if(self.with_att):
            [B, C, D, H, W] = list(x.shape)
            x = x.flatten(2).transpose(1, 2).contiguous()
            x = self.attn(x, D, H, W)[0]
            x = x.view(-1, D, H, W, C).permute(0, 4, 1, 2, 3).contiguous()
        return x

class AttDecoder(nn.Module):
    """
    Decoder of 3D UNet.

    Parameters are given in the `params` dictionary, and should include the
    following fields:

    :param in_chns: (int) Input channel number.
    :param feature_chns: (list) Feature channel for each resolution level. 
      The length should be 4 or 5, such as [16, 32, 64, 128, 256].
    :param dropout: (list) The dropout ratio for each resolution level. 
      The length should be the same as that of `feature_chns`.
    :param class_num: (int) The class number for segmentation task. 
    :param trilinear: (bool) Using bilinear for up-sampling or not. 
        If False, deconvolution will be used for up-sampling.
    """
    def __init__(self, params):
        super(AttDecoder, self).__init__()
        self.params    = params
        self.in_chns   = self.params['in_chns']
        self.ft_chns   = self.params['feature_chns']
        self.dropout   = self.params['dropout']
        self.n_class   = self.params['class_num']
        self.trilinear = self.params.get('trilinear', True)
        self.mul_pred  = self.params['multiscale_pred']
       
        assert(len(self.ft_chns) == 5 or len(self.ft_chns) == 4)

        if(len(self.ft_chns) == 5):
            self.up1 = AttUpBlock(self.ft_chns[4], self.ft_chns[3], self.ft_chns[3], self.dropout[3], self.trilinear) 
        att_params = {"input_resolution": [24, 24, 24], "depth": 2, "num_heads": 4}
        self.up2 = AttUpBlock(self.ft_chns[3], self.ft_chns[2], self.ft_chns[2], self.dropout[2], self.trilinear, True, att_params) 
        att_params = {"input_resolution": [48, 48, 48], "depth": 2, "num_heads": 4}
        self.up3 = AttUpBlock(self.ft_chns[2], self.ft_chns[1], self.ft_chns[1], self.dropout[1], self.trilinear, True, att_params) 
        self.up4 = AttUpBlock(self.ft_chns[1], self.ft_chns[0], self.ft_chns[0], self.dropout[0], self.trilinear) 
        self.out_conv = nn.Conv3d(self.ft_chns[0], self.n_class, kernel_size = 1)
        if(self.mul_pred):
            self.out_conv1 = nn.Conv3d(self.ft_chns[1], self.n_class, kernel_size = 1)
            self.out_conv2 = nn.Conv3d(self.ft_chns[2], self.n_class, kernel_size = 1)
            self.out_conv3 = nn.Conv3d(self.ft_chns[3], self.n_class, kernel_size = 1)

    def forward(self, x):
        if(len(self.ft_chns) == 5):
            assert(len(x) == 5)
            x0, x1, x2, x3, x4 = x 
            x_d3 = self.up1(x4, x3)
        else:
            assert(len(x) == 4)
            x0, x1, x2, x3 = x 
            x_d3 = x3
        x_d2 = self.up2(x_d3, x2)
        x_d1 = self.up3(x_d2, x1)
        x_d0 = self.up4(x_d1, x0)
        output = self.out_conv(x_d0)
        if(self.mul_pred):
            output1 = self.out_conv1(x_d1)
            output2 = self.out_conv2(x_d2)
            output3 = self.out_conv3(x_d3)
            output = [output, output1, output2, output3]
        return output

class MedFormerV2(nn.Module):
    """
    An implementation of the U-Net.
        
    * Reference: Özgün Çiçek, Ahmed Abdulkadir, Soeren S. Lienkamp, Thomas Brox, Olaf Ronneberger:
      3D U-Net: Learning Dense Volumetric Segmentation from Sparse Annotation. 
      `MICCAI (2) 2016: 424-432. <https://link.springer.com/chapter/10.1007/978-3-319-46723-8_49>`_
    
    Note that there are some modifications from the original paper, such as
    the use of batch normalization, dropout, leaky relu and deep supervision.

    Parameters are given in the `params` dictionary, and should include the
    following fields:

    :param in_chns: (int) Input channel number.
    :param feature_chns: (list) Feature channel for each resolution level. 
      The length should be 4 or 5, such as [16, 32, 64, 128, 256].
    :param dropout: (list) The dropout ratio for each resolution level. 
      The length should be the same as that of `feature_chns`.
    :param class_num: (int) The class number for segmentation task. 
    :param trilinear: (bool) Using trilinear for up-sampling or not. 
        If False, deconvolution will be used for up-sampling.
    """
    def __init__(self, params):
        super(MedFormerV2, self).__init__()
        self.params   = params
        self.encoder  = Encoder(params)
        self.decoder  = AttDecoder(params)  
        self.attn = Block(params)

    def forward(self, x):
        f = self.encoder(x)
        f[-1] = self.attn(f[-1])
        output = self.decoder(f)
        return output

if __name__ == "__main__":
    params = {'in_chns':4,
              'class_num': 2,
              'feature_chns':[16, 32, 64, 128],
              'dropout' : [0, 0, 0, 0.5],
              'trilinear': True,
              'multiscale_pred': True,
              'attention_hidden_size': 128,
              'attention_num_heads': 4,
              'attention_mlp_dim': 256,
              'attention_dropout_rate': 0.2}

    Net = MedFormerV2(params)
    Net = Net.double()

    x  = np.random.rand(1, 4, 96, 96, 96)
    xt = torch.from_numpy(x)
    xt = torch.tensor(xt)
    
    y = Net(xt)
    print("output length", len(y))
    for yi in y:
        yi = yi.detach().numpy()
        print(yi.shape)



# -*- coding: utf-8 -*-
from __future__ import print_function, division

import numpy as np 
import torch
import torch.nn as nn
from torch.nn.functional import interpolate
from pymic.net.net3d.unet3d import ConvBlock, Encoder
from pymic.net.net3d.trans3d.MedFormer_v1 import Block
from pymic.net.net3d.trans3d.MedFormer_v2 import SwinTransformerBlock, window_partition

class GLAttLayer(nn.Module):
    def __init__(self,
                 dim,
                 input_resolution,
                 num_heads,
                 window_size=7,
                 mlp_ratio=4.,
                 qkv_bias=True,
                 qk_scale=None,
                 drop=0.,
                 attn_drop=0.,
                 drop_path=0.,
                 norm_layer=nn.LayerNorm):
        super().__init__()
        self.window_size = window_size
        self.shift_size = window_size // 2
        # build blocks
        
        self.lcl_att = SwinTransformerBlock(
                dim=dim,
                input_resolution=input_resolution,
                num_heads=num_heads,
                window_size=window_size,
                shift_size=0,
                mlp_ratio=mlp_ratio,
                qkv_bias=qkv_bias,
                qk_scale=qk_scale,
                drop=drop,
                attn_drop=attn_drop,
                drop_path=drop_path)
        self.adpool = nn.AdaptiveAvgPool3d([12, 12, 12])

        params = {'attention_hidden_size': dim,
              'attention_num_heads': 4,
              'attention_mlp_dim': dim,
              'attention_dropout_rate': 0.2}
        self.glb_att = Block(params)
        self.conv1x1 = nn.Sequential(
            nn.Conv3d(2*dim, dim, kernel_size=1),
            nn.BatchNorm3d(dim),
            nn.LeakyReLU())

    def forward(self, x):
        [B, C, S, H, W] = list(x.shape)
        # calculate attention mask for SW-MSA
        Sp = int(np.ceil(S / self.window_size)) * self.window_size
        Hp = int(np.ceil(H / self.window_size)) * self.window_size
        Wp = int(np.ceil(W / self.window_size)) * self.window_size
        img_mask = torch.zeros((1, Sp, Hp, Wp, 1), device=x.device)  # 1 Hp Wp 1
        s_slices = (slice(0, -self.window_size),
                    slice(-self.window_size, -self.shift_size),
                    slice(-self.shift_size, None))
        h_slices = (slice(0, -self.window_size),
                    slice(-self.window_size, -self.shift_size),
                    slice(-self.shift_size, None))
        w_slices = (slice(0, -self.window_size),
                    slice(-self.window_size, -self.shift_size),
                    slice(-self.shift_size, None))
        cnt = 0
        for s in s_slices:
            for h in h_slices:
                for w in w_slices:
                    img_mask[:, s, h, w, :] = cnt
                    cnt += 1

        mask_windows = window_partition(img_mask, self.window_size)  
        mask_windows = mask_windows.view(-1,
                                         self.window_size * self.window_size * self.window_size)  
        attn_mask = mask_windows.unsqueeze(1) - mask_windows.unsqueeze(2)
        attn_mask = attn_mask.masked_fill(attn_mask != 0, float(-100.0)).masked_fill(attn_mask == 0, float(0.0))
        
        # for local attention
        xl = x.flatten(2).transpose(1, 2).contiguous()
        xl = self.lcl_att(xl, attn_mask)
        xl = xl.view(-1, S, H, W, C).permute(0, 4, 1, 2, 3).contiguous()

        # for global attention
        xg = self.adpool(x)
        xg = self.glb_att(xg)
        xg = interpolate(xg, [S, H, W], mode = 'trilinear')
        out = torch.cat([xl, xg], dim=1)
        out = self.conv1x1(out)
        return out 

class AttUpBlock(nn.Module):
    """
    3D upsampling followed by ConvBlock
    
    :param in_channels1: (int) Channel number of high-level features.
    :param in_channels2: (int) Channel number of low-level features.
    :param out_channels: (int) Output channel number.
    :param dropout_p: (int) Dropout probability.
    :param trilinear: (bool) Use trilinear for up-sampling (by default).
        If False, deconvolution is used for up-sampling. 
    """
    def __init__(self, in_channels1, in_channels2, out_channels, dropout_p,
                 trilinear=True, with_att = False, att_params = None):
        super(AttUpBlock, self).__init__()
        self.trilinear = trilinear
        self.with_att  = with_att
        if trilinear:
            self.conv1x1 = nn.Conv3d(in_channels1, in_channels2, kernel_size = 1)
            self.up = nn.Upsample(scale_factor=2, mode='trilinear', align_corners=True)
        else:
            self.up = nn.ConvTranspose3d(in_channels1, in_channels2, kernel_size=2, stride=2)
        self.conv = ConvBlock(in_channels2 * 2, out_channels, dropout_p)
        if(self.with_att):
            input_resolution = att_params['input_resolution']
            num_heads        = att_params['num_heads']
            window_size      = att_params['window_size']
            self.attn  = GLAttLayer(out_channels, input_resolution, num_heads, window_size, 2.0)

    def forward(self, x1, x2):
        if self.trilinear:
            x1 = self.conv1x1(x1)
        x1 = self.up(x1)
        x = torch.cat([x2, x1], dim=1)
        x =  self.conv(x) 
        if(self.with_att):
            x = self.attn(x)
        return x


class AttDecoder(nn.Module):
    """
    Decoder of 3D UNet.

    Parameters are given in the `params` dictionary, and should include the
    following fields:

    :param in_chns: (int) Input channel number.
    :param feature_chns: (list) Feature channel for each resolution level. 
      The length should be 4 or 5, such as [16, 32, 64, 128, 256].
    :param dropout: (list) The dropout ratio for each resolution level. 
      The length should be the same as that of `feature_chns`.
    :param class_num: (int) The class number for segmentation task. 
    :param trilinear: (bool) Using bilinear for up-sampling or not. 
        If False, deconvolution will be used for up-sampling.
    """
    def __init__(self, params):
        super(AttDecoder, self).__init__()
        self.params    = params
        self.in_chns   = self.params['in_chns']
        self.ft_chns   = self.params['feature_chns']
        self.dropout   = self.params['dropout']
        self.n_class   = self.params['class_num']
        self.trilinear = self.params.get('trilinear', True)
        self.mul_pred  = self.params['multiscale_pred']
       
        assert(len(self.ft_chns) == 5 or len(self.ft_chns) == 4)

        if(len(self.ft_chns) == 5):
            self.up1 = AttUpBlock(self.ft_chns[4], self.ft_chns[3], self.ft_chns[3], self.dropout[3], self.trilinear) 
        att_params = {"input_resolution": [24, 24, 24], "num_heads": 4, "window_size": 7}
        self.up2 = AttUpBlock(self.ft_chns[3], self.ft_chns[2], self.ft_chns[2], self.dropout[2], self.trilinear, True, att_params) 
        att_params = {"input_resolution": [48, 48, 48], "num_heads": 4, "window_size": 7}
        self.up3 = AttUpBlock(self.ft_chns[2], self.ft_chns[1], self.ft_chns[1], self.dropout[1], self.trilinear, True, att_params) 
        self.up4 = AttUpBlock(self.ft_chns[1], self.ft_chns[0], self.ft_chns[0], self.dropout[0], self.trilinear) 
        self.out_conv = nn.Conv3d(self.ft_chns[0], self.n_class, kernel_size = 1)
        if(self.mul_pred):
            self.out_conv1 = nn.Conv3d(self.ft_chns[1], self.n_class, kernel_size = 1)
            self.out_conv2 = nn.Conv3d(self.ft_chns[2], self.n_class, kernel_size = 1)
            self.out_conv3 = nn.Conv3d(self.ft_chns[3], self.n_class, kernel_size = 1)

    def forward(self, x):
        if(len(self.ft_chns) == 5):
            assert(len(x) == 5)
            x0, x1, x2, x3, x4 = x 
            x_d3 = self.up1(x4, x3)
        else:
            assert(len(x) == 4)
            x0, x1, x2, x3 = x 
            x_d3 = x3
        x_d2 = self.up2(x_d3, x2)
        x_d1 = self.up3(x_d2, x1)
        x_d0 = self.up4(x_d1, x0)
        output = self.out_conv(x_d0)
        if(self.mul_pred):
            output1 = self.out_conv1(x_d1)
            output2 = self.out_conv2(x_d2)
            output3 = self.out_conv3(x_d3)
            output = [output, output1, output2, output3]
        return output

class MedFormerV3(nn.Module):
    """
    An implementation of the U-Net.
        
    * Reference: Özgün Çiçek, Ahmed Abdulkadir, Soeren S. Lienkamp, Thomas Brox, Olaf Ronneberger:
      3D U-Net: Learning Dense Volumetric Segmentation from Sparse Annotation. 
      `MICCAI (2) 2016: 424-432. <https://link.springer.com/chapter/10.1007/978-3-319-46723-8_49>`_
    
    Note that there are some modifications from the original paper, such as
    the use of batch normalization, dropout, leaky relu and deep supervision.

    Parameters are given in the `params` dictionary, and should include the
    following fields:

    :param in_chns: (int) Input channel number.
    :param feature_chns: (list) Feature channel for each resolution level. 
      The length should be 4 or 5, such as [16, 32, 64, 128, 256].
    :param dropout: (list) The dropout ratio for each resolution level. 
      The length should be the same as that of `feature_chns`.
    :param class_num: (int) The class number for segmentation task. 
    :param trilinear: (bool) Using trilinear for up-sampling or not. 
        If False, deconvolution will be used for up-sampling.
    """
    def __init__(self, params):
        super(MedFormerV3, self).__init__()
        self.params   = params
        self.encoder  = Encoder(params)
        self.decoder  = AttDecoder(params)  
        params["attention_hidden_size"] = params['feature_chns'][-1]
        params["attention_mlp_dim"]     = params['feature_chns'][-1]
        self.attn = Block(params)

    def forward(self, x):
        f = self.encoder(x)
        f[-1] = self.attn(f[-1])
        output = self.decoder(f)
        return output

if __name__ == "__main__":
    params = {'in_chns':4,
              'class_num': 2,
              'feature_chns':[16, 32, 64, 128],
              'dropout' : [0, 0, 0, 0.5],
              'trilinear': True,
              'multiscale_pred': True,
              'attention_num_heads': 4,
              'attention_dropout_rate': 0.2}

    Net = MedFormerV3(params)
    Net = Net.double()

    x  = np.random.rand(2, 4, 96, 96, 96)
    xt = torch.from_numpy(x)
    xt = torch.tensor(xt)
    
    y = Net(xt)
    print("output length", len(y))
    for yi in y:
        yi = yi.detach().numpy()
        print(yi.shape)



# -*- coding: utf-8 -*-
from __future__ import print_function, division

import torch
from torch import optim
from torch.optim import lr_scheduler
from pymic.util.general import keyword_match

def get_optimizer(name, net_params, optim_params):
    lr = optim_params['learning_rate']
    momentum = optim_params['momentum']
    weight_decay = optim_params['weight_decay']
    # see https://www.codeleading.com/article/44815584159/
    param_group = [{'params': net_params, 'initial_lr': lr}]
    if(keyword_match(name, "SGD")):
        return optim.SGD(param_group, lr, 
            momentum = momentum, weight_decay = weight_decay)
    elif(keyword_match(name, "Adam")):
        return optim.Adam(param_group, lr, weight_decay = weight_decay)
    elif(keyword_match(name, "SparseAdam")):
        return optim.SparseAdam(param_group, lr)
    elif(keyword_match(name, "Adadelta")):
        return optim.Adadelta(param_group, lr, weight_decay = weight_decay)
    elif(keyword_match(name, "Adagrad")): 
        return optim.Adagrad(param_group, lr, weight_decay = weight_decay)
    elif(keyword_match(name, "Adamax")): 
        return optim.Adamax(param_group, lr, weight_decay = weight_decay)
    elif(keyword_match(name, "ASGD")): 
        return optim.ASGD(param_group, lr, weight_decay = weight_decay)
    elif(keyword_match(name, "LBFGS")): 
        return optim.LBFGS(param_group, lr)
    elif(keyword_match(name, "RMSprop")): 
        return optim.RMSprop(param_group, lr, momentum = momentum,
            weight_decay = weight_decay)
    elif(keyword_match(name, "Rprop")): 
        return optim.Rprop(param_group, lr)
    else:
        raise ValueError("unsupported optimizer {0:}".format(name))


def get_lr_scheduler(optimizer, sched_params):
    name       = sched_params["lr_scheduler"]
    val_it     = sched_params["iter_valid"]
    epoch_last = sched_params["last_iter"] 
    if(epoch_last > 0):
        epoch_last = int(epoch_last / val_it)
    if(name is None):
        return None
    if(keyword_match(name, "ReduceLROnPlateau")):
        patience_it = sched_params["ReduceLROnPlateau_patience".lower()]
        patience = patience_it / val_it   
        lr_gamma = sched_params["lr_gamma"]
        scheduler = lr_scheduler.ReduceLROnPlateau(optimizer,
            mode = "max", factor=lr_gamma, patience = patience)
    elif(keyword_match(name, "MultiStepLR")):
        lr_milestones = sched_params["lr_milestones"]
        lr_milestones = [int(item / val_it) for item in lr_milestones]
        lr_gamma  = sched_params["lr_gamma"]
        scheduler = lr_scheduler.MultiStepLR(optimizer,
                    lr_milestones, lr_gamma, epoch_last)
    elif(keyword_match(name, "StepLR")):
        lr_step   = sched_params["lr_step"] / val_it
        lr_gamma  = sched_params["lr_gamma"]
        scheduler = lr_scheduler.StepLR(optimizer,
                    lr_step, lr_gamma, epoch_last)
    elif(keyword_match(name, "CosineAnnealingLR")):
        epoch_max  = sched_params["iter_max"] / val_it
        lr_min     = sched_params.get("lr_min", 0)
        scheduler = lr_scheduler.CosineAnnealingLR(optimizer,
                    epoch_max, lr_min, epoch_last)
    elif(keyword_match(name, "PolynomialLR")):
        epoch_max  = sched_params["iter_max"] / val_it
        power      = sched_params["lr_power"]
        scheduler = lr_scheduler.PolynomialLR(optimizer,
                    epoch_max, power, epoch_last)
    else:
        raise ValueError("unsupported lr scheduler {0:}".format(name))
    return scheduler


# -*- coding: utf-8 -*-
"""
Evaluation module for classification tasks.
"""
from __future__ import absolute_import, print_function

import os
import csv 
import sys
import math
import pandas as pd
import random
import configparser
import numpy as np
import sklearn.metrics as metrics
from scipy import ndimage
from pymic.io.image_read_write import *
from pymic.util.image_process import *
from pymic.util.parse_config import parse_config

def accuracy(gt_label, pred_label):
    """
    Calculate the accuracy. 
    """
    correct_pred = gt_label == pred_label
    acc = (correct_pred.sum() + 0.0 ) / len(gt_label)
    return acc 

def sensitivity(gt_label, pred_label):
    """
    Calculate the sensitivity for binary prediction. 
    """
    pos_pred = gt_label * pred_label
    senst = (pos_pred.sum() + 0.0) / gt_label.sum() 
    return senst

def specificity(gt_label, pred_label):
    """
    Calculate the specificity for binary prediction. 
    """
    gt_label = 1 - gt_label
    pred_label = 1 - pred_label
    neg_pred = gt_label * pred_label
    spec = (neg_pred.sum() + 0.0) / gt_label.sum() 
    return spec 

def get_evaluation_score(gt_label, pred_prob, metric):
    """
    Get an evaluation score for binary classification.

    :param gt_label: (array) Ground truth label.
    :param pred_prob: (array) Predicted positive probability. 
    :param metric: (str) One of the evaluation metrics in 
        {`accuracy`, `recall`, `sensitivity`, `specificity`,
        `precision`, `auc`}.
    """
    pred_lab = np.argmax(pred_prob, axis = 1)
    if(metric == "accuracy"):
        score = metrics.accuracy_score(gt_label, pred_lab)
    elif(metric == "recall" or metric == "sensitivity"):
        score = metrics.recall_score(gt_label, pred_lab)
    elif(metric == "specificity"):
        score = metrics.recall_score(1 - gt_label, 1 - pred_lab)
    elif(metric == "precision"):
        score = metrics.precision_score(gt_label, pred_lab)
    elif(metric == "auc"):
        score = metrics.roc_auc_score(gt_label, pred_prob[:,1])
    else:
        raise ValueError("undefined metric: {0:}".format(metric))
    return score

def binary_evaluation(config):
    """
    Evaluation of binary classification performance.
    The arguments are given in the `config` dictionary. 
    It should have the following fields:

    :param metric_list: (list) A list of evaluation metrics.
        The supported metrics are {`accuracy`, `recall`, `sensitivity`, `specificity`,
        `precision`, `auc`}.
    :param ground_truth_csv: (str) The csv file for ground truth.
    :param predict_prob_csv: (str) The csv file for prediction probability.
    """
    metric_list = config['metric_list']
    gt_csv  = config['ground_truth_csv']
    prob_csv= config['predict_prob_csv']
    gt_items  = pd.read_csv(gt_csv)
    prob_items = pd.read_csv(prob_csv)
    assert(len(gt_items) == len(prob_items))
    for i in range(len(gt_items)):
        assert(gt_items.iloc[i, 0] == prob_items.iloc[i, 0])
    
    gt_data   = np.asarray(gt_items.iloc[:, -1])
    prob_data = np.asarray(prob_items.iloc[:, 1:])
    score_list = []
    for metric in metric_list:
        score = get_evaluation_score(gt_data, prob_data, metric)
        score_list.append(score)
        print("{0:}: {1:}".format(metric, score))

    out_csv = prob_csv.replace("prob", "eval")
    with open(out_csv, mode='w') as csv_file:
        csv_writer = csv.writer(csv_file, delimiter=',', 
                            quotechar='"',quoting=csv.QUOTE_MINIMAL)
        csv_writer.writerow(metric_list)
        csv_writer.writerow(score_list)

def nexcl_evaluation(config):
    """
    Evaluation of non-exclusive binary classification performance.
    The arguments are given in the `config` dictionary. 
    It should have the following fields:

    :param metric_list: (list) A list of evaluation metrics.
        The supported metrics are {`accuracy`, `recall`, `sensitivity`, `specificity`,
        `precision`, `auc`}.
    :param ground_truth_csv: (str) The csv file for ground truth.
    :param predict_prob_csv: (str) The csv file for prediction probability.
    """
    metric_list = config['metric_list']
    gt_csv    = config['ground_truth_csv']
    prob_csv  = config['predict_prob_csv']
    gt_items  = pd.read_csv(gt_csv)
    prob_items= pd.read_csv(prob_csv)
    assert(len(gt_items) == len(prob_items))
    for i in range(len(gt_items)):
        assert(gt_items.iloc[i, 0] == prob_items.iloc[i, 0])
    
    cls_names = gt_items.columns[1:]
    cls_num   = len(cls_names)
    gt_data  = np.asarray(gt_items.iloc[:, 1:cls_num + 1])
    prob_data = np.asarray(prob_items.iloc[:, 1:cls_num + 1])
    score_list= []
    for metric in metric_list:
        print(metric)
        score_m = []
        for c in range(cls_num):
            gt_data_c = gt_data[:, c:c+1]
            prob_c = prob_data[:, c]
            prob_c = np.asarray([1.0 - prob_c, prob_c])
            prob_c = np.transpose(prob_c)
            score = get_evaluation_score(gt_data_c, prob_c, metric)
            score_m.append(score)
            print(cls_names[c], score)
        score_avg = np.asarray(score_m).mean()
        print('avg', score_avg)
        score_m.append(score_avg)
        score_list.append(score_m)

    out_csv = prob_csv.replace("prob", "eval")
    with open(out_csv, mode='w') as csv_file:
        csv_writer = csv.writer(csv_file, delimiter=',', 
                            quotechar='"',quoting=csv.QUOTE_MINIMAL)
        csv_writer.writerow(['metric'] + list(cls_names) + ['avg'])
        for i in range(len(score_list)):
            item = metric_list[i : i+1] + score_list[i]
            csv_writer.writerow(item)

def main():
    """
    Main function for evaluation of classification results. 
    A configuration file is needed for runing. e.g., 
    
    .. code-block:: none

        pymic_evaluate_cls config.cfg

    The configuration file should have an `evaluation` section with
    the following fields:

    :param task_type: (str) `cls` or `cls_nexcl`.
    :param metric_list: (list) A list of evaluation metrics.
        The supported metrics are {`accuracy`, `recall`, `sensitivity`, `specificity`,
        `precision`, `auc`}.
    :param ground_truth_csv: (str) The csv file for ground truth.
    :param predict_prob_csv: (str) The csv file for prediction probability.
    """
    if(len(sys.argv) < 2):
        print('Number of arguments should be 2. e.g.')
        print('    pymic_evaluate_cls config.cfg')
        exit()
    config_file = str(sys.argv[1])
    assert(os.path.isfile(config_file))
    config = parse_config(config_file)['evaluation']
    task_type = config.get('task_type', "cls")
    if(task_type == "cls"):  # default exclusive classification
        binary_evaluation(config)
    else:                    # non exclusive classification
        nexcl_evaluation(config)
    
if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
The code is adapted from the original implementation at `Github. 
<https://github.com/LEONOB2014/GatedCRFLoss/blob/master/models/model_loss_semseg_gatedcrf.py>`_
"""
import torch
import torch.nn.functional as F

class GatedCRFLoss(torch.nn.Module):
    """
    Gated CRF Loss for Weakly Supervised Semantic Image Segmentation.
    This loss function promotes consistent label assignment guided by input features, such as RGBXY.
    
    * Reference: Anton Obukhov, Stamatios Georgoulis, Dengxin Dai and Luc Van Gool: 
      Gated CRF Loss for Weakly Supervised Semantic Image Segmentation. `CoRR 
      <http://arxiv.org/abs/1906.04651>`_ 2019.
    """
    def forward(
            self, y_hat_softmax, kernels_desc, kernels_radius, sample, height_input, width_input,
            mask_src=None, mask_dst=None, compatibility=None, custom_modality_downsamplers=None, out_kernels_vis=False
    ):
        """
        Performs the forward pass of the loss.

        :param y_hat_softmax: A tensor of predicted per-pixel class probabilities of size NxCxHxW
        :param kernels_desc: A list of dictionaries, each describing one Gaussian kernel composition from modalities.
            The final kernel is a weighted sum of individual kernels. Following example is a composition of
            RGBXY and XY kernels:
            kernels_desc: [{'weight': 0.9,'xy': 6,'rgb': 0.1},{'weight': 0.1,'xy': 6}]

        :param kernels_radius: Defines size of bounding box region around each pixel in which the kernel is constructed.
        :param sample: A dictionary with modalities (except 'xy') used in kernels_desc parameter. Each of the provided
            modalities is allowed to be larger than the shape of y_hat_softmax, in such case downsampling will be
            invoked. Default downsampling method is area resize; this can be overriden by setting.
            custom_modality_downsamplers parameter.
        :param width_input, height_input: Dimensions of the full scale resolution of modalities
        :param mask_src: (optional) Source mask.
        :param mask_dst: (optional) Destination mask.
        :param compatibility: (optional) Classes compatibility matrix, defaults to Potts model.
        :param custom_modality_downsamplers: A dictionary of modality downsampling functions.
        :param out_kernels_vis: Whether to return a tensor with kernels visualized with some step.
        :return: Loss function value.
        """
        assert y_hat_softmax.dim() == 4, 'Prediction must be a NCHW batch'
        N, C, height_pred, width_pred = y_hat_softmax.shape
        device = y_hat_softmax.device

        assert width_input % width_pred == 0 and height_input % height_pred == 0 and \
               width_input * height_pred == height_input * width_pred, \
            f'[{width_input}x{height_input}] !~= [{width_pred}x{height_pred}]'

        kernels = self._create_kernels(
            kernels_desc, kernels_radius, sample, N, height_pred, width_pred, device, custom_modality_downsamplers
        )

        denom = N * height_pred * width_pred

        def resize_fix_mask(mask, name):
            assert mask.dim() == 4 and mask.shape[:2] == (N, 1) and mask.dtype == torch.float32, \
                f'{name} mask must be a NCHW batch with C=1 and dtype float32'
            if mask.shape[2:] != (height_pred, width_pred):
                mask = GatedCRFLoss._downsample(
                    mask, 'mask', height_pred, width_pred, custom_modality_downsamplers
                )
            mask[mask != mask] = 0.0    # handle NaN
            mask[mask < 1.0] = 0.0      # handle edges of mask after interpolation
            return mask

        if mask_src is not None:
            mask_src = resize_fix_mask(mask_src, 'Source')
            denom = mask_src.sum().clamp(min=1)
            mask_src = self._unfold(mask_src, kernels_radius)
            kernels = kernels * mask_src

        if mask_dst is not None:
            mask_dst = resize_fix_mask(mask_dst, 'Destination')
            denom = mask_dst.sum().clamp(min=1)
            mask_dst = mask_dst.view(N, 1, 1, 1, height_pred, width_pred)
            kernels = kernels * mask_dst

        y_hat_unfolded = self._unfold(y_hat_softmax, kernels_radius)

        product_kernel_x_y_hat = (kernels * y_hat_unfolded) \
            .view(N, C, (kernels_radius * 2 + 1) ** 2, height_pred, width_pred) \
            .sum(dim=2, keepdim=False)

        if compatibility is None:
            # Using shortcut for Pott's class compatibility model
            loss = -(product_kernel_x_y_hat * y_hat_softmax).sum()
            loss = kernels.sum() + loss  # comment out to save computation, total loss may go below 0
        else:
            assert compatibility.shape == (C, C), f'Compatibility matrix expected shape [{C}x{C}]'
            assert (compatibility < 0).int().sum() == 0, 'Compatibility matrix must not have negative values'
            assert compatibility.diag.sum() == 0, 'Compatibility matrix diagonal must be 0'
            compat = (C-1) * F.normalize(compatibility.float().to(device), p=1, dim=1)
            y_hat_CxNHW = y_hat_softmax.permute(1, 0, 2, 3).contiguous().view(C, -1)
            product_kernel_x_y_hat_NHWxC = product_kernel_x_y_hat.permute(0, 2, 3, 1).contiguous().view(-1, C)
            product_CxC = torch.mm(y_hat_CxNHW, product_kernel_x_y_hat_NHWxC)
            loss = (compat * product_CxC).sum()

        out = {
            'loss': loss / denom,
        }

        if out_kernels_vis:
            out['kernels_vis'] = self._visualize_kernels(
                kernels, kernels_radius, height_input, width_input, height_pred, width_pred
            )

        return out

    @staticmethod
    def _downsample(img, modality, height_dst, width_dst, custom_modality_downsamplers):
        if custom_modality_downsamplers is not None and modality in custom_modality_downsamplers:
            f_down = custom_modality_downsamplers[modality]
        else:
            f_down = F.adaptive_avg_pool2d
        return f_down(img, (height_dst, width_dst))

    @staticmethod
    def _create_kernels(
            kernels_desc, kernels_radius, sample, N, height_pred, width_pred, device, custom_modality_downsamplers
    ):
        kernels = None
        for i, desc in enumerate(kernels_desc):
            weight = desc['weight']
            features = []
            for modality, sigma in desc.items():
                if modality == 'weight':
                    continue
                if modality == 'xy':
                    feature = GatedCRFLoss._get_mesh(N, height_pred, width_pred, device)
                else:
                    assert modality in sample, \
                        f'Modality {modality} is listed in {i}-th kernel descriptor, but not present in the sample'
                    feature = sample[modality]
                    feature = GatedCRFLoss._downsample(
                        feature, modality, height_pred, width_pred, custom_modality_downsamplers
                    )
                feature /= sigma
                features.append(feature)
            features = torch.cat(features, dim=1)
            kernel = weight * GatedCRFLoss._create_kernels_from_features(features, kernels_radius)
            kernels = kernel if kernels is None else kernel + kernels
        return kernels

    @staticmethod
    def _create_kernels_from_features(features, radius):
        assert features.dim() == 4, 'Features must be a NCHW batch'
        N, C, H, W = features.shape
        kernels = GatedCRFLoss._unfold(features, radius)
        kernels = kernels - kernels[:, :, radius, radius, :, :].view(N, C, 1, 1, H, W)
        kernels = (-0.5 * kernels ** 2).sum(dim=1, keepdim=True).exp()
        kernels[:, :, radius, radius, :, :] = 0
        return kernels

    @staticmethod
    def _get_mesh(N, H, W, device):
        return torch.cat((
            torch.arange(0, W, 1, dtype=torch.float32, device=device).view(1, 1, 1, W).repeat(N, 1, H, 1),
            torch.arange(0, H, 1, dtype=torch.float32, device=device).view(1, 1, H, 1).repeat(N, 1, 1, W)
        ), 1)

    @staticmethod
    def _unfold(img, radius):
        assert img.dim() == 4, 'Unfolding requires NCHW batch'
        N, C, H, W = img.shape
        diameter = 2 * radius + 1
        return F.unfold(img, diameter, 1, radius).view(N, C, diameter, diameter, H, W)

    @staticmethod
    def _visualize_kernels(kernels, radius, height_input, width_input, height_pred, width_pred):
        diameter = 2 * radius + 1
        vis = kernels[:, :, :, :, radius::diameter, radius::diameter]
        vis_nh, vis_nw = vis.shape[-2:]
        vis = vis.permute(0, 1, 4, 2, 5, 3).contiguous().view(kernels.shape[0], 1, diameter * vis_nh, diameter * vis_nw)
        if vis.shape[2] > height_pred:
            vis = vis[:, :, :height_pred, :]
        if vis.shape[3] > width_pred:
            vis = vis[:, :, :, :width_pred]
        if vis.shape[2:] != (height_pred, width_pred):
            vis = F.pad(vis, [0, width_pred-vis.shape[3], 0, height_pred-vis.shape[2]])
        vis = F.interpolate(vis, (height_input, width_input), mode='nearest')
        return vis



# -*- coding: utf-8 -*-
from __future__ import print_function, division

import math
import torch
import torch.nn as nn
import numpy as np
from torch.nn import Dropout, Softmax, Linear, Conv2d, LayerNorm
from pymic.net.net3d.unet3d import Decoder

class EmbeddingBlock(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size, padding, stride):
        super(EmbeddingBlock, self).__init__()
        self.out_channels = out_channels
        self.conv1 = nn.Conv3d(in_channels, out_channels//2, kernel_size=kernel_size, padding=padding, stride = stride)
        self.conv2 = nn.Conv3d(out_channels//2, out_channels, kernel_size=1)
        self.act  = nn.GELU()
        self.norm1 = nn.LayerNorm(out_channels//2)
        self.norm2 = nn.LayerNorm(out_channels)
        
       
    def forward(self, x):
        x = self.act(self.conv1(x))
        # norm 1
        Ws, Wh, Ww = x.size(2), x.size(3), x.size(4)
        x = x.flatten(2).transpose(1, 2).contiguous()
        x = self.norm1(x)
        x = x.transpose(1, 2).contiguous().view(-1, self.out_channels // 2, Ws, Wh, Ww)

        x = self.act(self.conv2(x))
        x = x.flatten(2).transpose(1, 2).contiguous()
        x = self.norm2(x)
        x = x.transpose(1, 2).contiguous().view(-1, self.out_channels, Ws, Wh, Ww)

        return x

class Encoder(nn.Module):
    """
    Encoder of 3D UNet.

    Parameters are given in the `params` dictionary, and should include the
    following fields:

    :param in_chns: (int) Input channel number.
    :param feature_chns: (list) Feature channel for each resolution level. 
      The length should be 4 or 5, such as [16, 32, 64, 128, 256].
    :param dropout: (list) The dropout ratio for each resolution level. 
      The length should be the same as that of `feature_chns`.
    """
    def __init__(self, params):
        super(Encoder, self).__init__()
        self.params    = params
        self.in_chns   = self.params['in_chns']
        self.ft_chns   = self.params['feature_chns']
        assert(len(self.ft_chns) == 4)

        self.down0  = EmbeddingBlock(self.in_chns, self.ft_chns[0], 3, 1, 1)
        self.down1  = EmbeddingBlock(self.in_chns, self.ft_chns[1], 2, 0, 2)
        self.down2  = EmbeddingBlock(self.in_chns, self.ft_chns[2], 4, 0, 4)
        self.down3  = EmbeddingBlock(self.in_chns, self.ft_chns[3], 8, 0, 8)
        
    def forward(self, x):
        x0 = self.down0(x)
        x1 = self.down1(x)
        x2 = self.down2(x)
        x3 = self.down3(x)
        output = [x0, x1, x2, x3]
        return output

class MedFormerVA1(nn.Module):
    def __init__(self, params):
        super(MedFormerVA1, self).__init__()
        self.params   = params
        self.encoder  = Encoder(params)
        self.decoder  = Decoder(params)  

    def forward(self, x):
        f = self.encoder(x)
        output = self.decoder(f)
        return output


if __name__ == "__main__":
    params = {'in_chns':1,
              'class_num': 8,
              'feature_chns':[16, 32, 64, 128],
              'dropout' : [0, 0, 0, 0.5],
              'trilinear': True,
              'deep_supervise': True,
              'attention_hidden_size': 128,
              'attention_num_heads': 4,
              'attention_mlp_dim': 256,
              'attention_dropout_rate': 0.2}
    Net = MedFormerVA1(params)
    Net = Net.double()

    x  = np.random.rand(1, 1, 96, 96, 96)
    xt = torch.from_numpy(x)
    xt = torch.tensor(xt)
    
    y = Net(xt)
    print("output length", len(y))
    for yi in y:
        yi = yi.detach().numpy()
        print(yi.shape)


# -*- coding: utf-8 -*-
from __future__ import print_function, division

import torch
import torch.nn as nn

class AbstractSegLoss(nn.Module):
    """
    Abstract class for loss function of segmentation tasks.
    The parameters should be written in the `params` dictionary, and it has the
    following fields:

    :param `loss_softmax`: (optional, bool) 
        Apply softmax to the prediction of network or not. Default is True.
    """
    def __init__(self, params = None):
        super(AbstractSegLoss, self).__init__()
        if(params is None):
            self.softmax = True
        else:
            self.softmax = params.get('loss_softmax', True)

    def forward(self, loss_input_dict):
        """
        Forward pass for calculating the loss.
        The arguments should be written in the `loss_input_dict` dictionary, 
        and it has the following fields:

        :param `prediction`: (tensor) Prediction of a network, with the 
            shape of [N, C, D, H, W] or [N, C, H, W].
        :param `ground_truth`: (tensor) Ground truth, with the 
            shape of [N, C, D, H, W] or [N, C, H, W]. 
        :param `pixel_weight`: (optional) Pixel-wise weight map, with the
            shape of [N, 1, D, H, W] or [N, 1, H, W]. Default is None.
        :return: Loss function value.
        """
        pass



# -*- coding: utf-8 -*-

from __future__ import print_function, division

import torch
import torch.nn as nn
from pymic.loss.seg.abstract import AbstractSegLoss
from pymic.loss.seg.util import reshape_tensor_to_2D

class SLSRLoss(AbstractSegLoss):
    """
    Spatial Label Smoothing Regularization (SLSR) loss for learning from
    noisy annotatins. This loss requires pixel weighting, please make sure
    that a `pixel_weight` field is provided for the csv file of the training images.

    The pixel wight here is actually the confidence mask, i.e., if the value is one, 
    it means the label of corresponding pixel is noisy and should be smoothed.

    * Reference: Minqing Zhang, Jiantao Gao et al.: Characterizing Label Errors: Confident Learning for Noisy-Labeled Image 
      Segmentation, `MICCAI 2020. <https://link.springer.com/chapter/10.1007/978-3-030-59710-8_70>`_ 
    
    The arguments should be written in the `params` dictionary, and it has the
    following fields:

    :param `loss_softmax`: (bool) Apply softmax to the prediction of network or not. 
    :param `slsrloss_epsilon`: (optional, float) Hyper-parameter epsilon. Default is 0.25.
    """
    def __init__(self, params = None):
        super(SLSRLoss, self).__init__(params)
        if(params is None):
            params = {}
        self.epsilon = params.get('slsrloss_epsilon', 0.25)
    
    def forward(self, loss_input_dict):
        predict = loss_input_dict['prediction']
        soft_y  = loss_input_dict['ground_truth']
        pix_w   = loss_input_dict.get('pixel_weight', None)

        if(isinstance(predict, (list, tuple))):
            predict = predict[0]
        if(self.softmax):
            predict = nn.Softmax(dim = 1)(predict)
        predict = reshape_tensor_to_2D(predict)
        soft_y  = reshape_tensor_to_2D(soft_y)
        if(pix_w is not None):
            pix_w   = reshape_tensor_to_2D(pix_w > 0).float()
            # smooth labels for pixels in the unconfident mask 
            smooth_y = (soft_y - 0.5) * (0.5 - self.epsilon) / 0.5 + 0.5
            smooth_y = pix_w * smooth_y + (1 - pix_w) * soft_y
        else:
            smooth_y = soft_y

        # for numeric stability
        predict = predict * 0.999 + 5e-4
        ce = - smooth_y* torch.log(predict)
        ce = torch.sum(ce, dim = 1) # shape is [N]
        ce = torch.mean(ce)  
        return ce



# -*- coding: utf-8 -*-
from __future__ import print_function, division

import json
import random
import numpy as np
from pymic.transform.abstract_transform import AbstractTransform


class RandomTranspose(AbstractTransform):
    """
    Random transpose for 3D volumes. Assume the input has a shape of [C, D, H, W], the
    output shape will be of [C, D, H, W], [C, W, H, D] or [C, H, D, W]

    The arguments should be written in the `params` dictionary, and it has the
    following fields:

    :param `RandomTranspose_inverse`: (optional, bool) 
        Is inverse transform needed for inference. Default is `True`.
    """
    def __init__(self, params): 
        super(RandomTranspose, self).__init__(params)
        self.inverse = params.get('RandomTranspose_inverse'.lower(), True)

    def __call__(self, sample):
        image = sample['image']
        input_shape = image.shape
        input_dim = len(input_shape) - 1
        assert(input_dim == 3)

        rand_num = random.random()
        if(rand_num < 0.4):
            transpose_axis = None 
        elif(rand_num < 0.7):
            transpose_axis = [0, 3, 2, 1]
        else:
            transpose_axis = [0, 2, 1, 3]
        sample['RandomTranspose_Param'] = json.dumps(transpose_axis) 
        if(transpose_axis is not None):
            image_t = np.transpose(image, transpose_axis)
            sample['image'] = image_t
            if('label' in sample and self.task == 'segmentation'):
                sample['label'] = np.transpose(sample['label'] , transpose_axis)
            if('pixel_weight' in sample and self.task == 'segmentation'):
                sample['pixel_weight'] = np.transpose(sample['pixel_weight'] , transpose_axis)

            
        return sample

    def  inverse_transform_for_prediction(self, sample):
        if(isinstance(sample['RandomTranspose_Param'], list) or \
            isinstance(sample['RandomTranspose_Param'], tuple)):
            transpose_axis = json.loads(sample['RandomTranspose_Param'][0]) 
        else:
            transpose_axis = json.loads(sample['RandomTranspose_Param'])
        if(transpose_axis is not None):
            sample['predict']  = np.transpose(sample['predict'] , transpose_axis)
        return sample